/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "bdbtripletbtreecursor.h"
#include "rdfsbdbtriplet.h"

#include <stdlib.h>

using namespace Rdfs;

BdbTripleTBtreeCursor::BdbTripleTBtreeCursor(Rdfs::RdfsBdbTripleT* backend, const Rdfs::Triple& query, Rdfs::RdfsRole role, Rdfs::RdfsTable table)
    : RdfsCursor(backend, query, table)
{
    // Subject contains encoded (prefixed) rule
    m_db = backend->btreeDatabase(query.subject, role, table);
    m_primary = query.subject;
    memset(&m_current, 0, sizeof(DBT));
    m_cursor = NULL;
    m_eof = true;
    m_open = false;
}

BdbTripleTBtreeCursor::~BdbTripleTBtreeCursor()
{
    close();
}


bool BdbTripleTBtreeCursor::eof() const
{
    return m_eof;
}

void BdbTripleTBtreeCursor::close()
{
    if (m_cursor != NULL) {
        m_cursor->c_close(m_cursor);
        m_cursor = NULL;
    }

    m_eof = true;
    m_open = false;
}

bool BdbTripleTBtreeCursor::open()
{
    m_cursor = loadCursor(queryTriple());
    if (m_cursor == NULL) {
        m_eof = true;
        m_open = false;
        return false;
    }

    // Load current key
    DBT data;
    memset(&data, 0, sizeof(DBT));
    m_cursor->c_get(m_cursor, &m_current, &data, DB_CURRENT);

    m_currentTriple = RdfsBdbTripleT::btreeTripleFromDbt(&m_current, &data, m_primary);

    m_eof = false;
    moveValid();

    // Check if we have a result
    m_open = !eof();
    if (!m_open) {
        close();
    }

    return m_open;
}

bool BdbTripleTBtreeCursor::next()
{
    advanceCursor();
    return moveValid();
}

Rdfs::Triple BdbTripleTBtreeCursor::current() const
{
    return m_currentTriple;
}

DBC* BdbTripleTBtreeCursor::loadCursor(const Rdfs::Triple& t) const
{
    DB *db = m_db;
    DBT key = RdfsBdbTripleT::btreeDbt(t.predicate, t.object);
    DBT data;
    memset(&data, 0, sizeof(DBT));

    DBC* cursor;

    int ret;
    u_int32_t flags, cflags;
    flags = cflags = 0;
#ifdef DB_BULK
    // Add BULK cursor to prevent unnecessary cache misses
    flags = DB_BULK;
#endif

    // Init cursor
    if ((ret = db->cursor(db, NULL, &cursor, flags)) != 0) {
        free(key.data);
        return NULL;
    }
    void *dataptr = key.data;

    // If one atom is empty: do range traversal
    cflags = (t.predicate.isEmpty() || t.object.isEmpty())
           ? DB_SET_RANGE
           : DB_SET;
    // Set cursor to correct position
    ret = cursor->c_get(cursor, &key, &data, cflags);
    // Free allocated memory - the key has been overwritten
    free(dataptr);

    if (ret == 0) {
        return cursor;
    }
    else {
        cursor->c_close(cursor);
        return NULL;
    }
}

bool BdbTripleTBtreeCursor::isMatch() const
{
    Triple q = queryTriple();
    return ((q.predicate.isEmpty() || q.predicate == m_currentTriple.predicate) &&
            (q.object.isEmpty() || q.object == m_currentTriple.object));
}

bool BdbTripleTBtreeCursor::moveValid()
{
    if (eof()) {
        return false;
    }

    while (!isMatch()) {
        // Check if advancing cursor may yield new triple
        if (!mayHaveMore()) {
            m_eof = true; // Found all matches
            return false;
        }

        // Move on
        advanceCursor();

        if (eof()) {
            return false;
        }
    }
    // Found a match
    return true;
}


bool BdbTripleTBtreeCursor::mayHaveMore()
{
    QString queryStrings[] = { queryTriple().predicate, queryTriple().object };
    QString currentStrings[] = { m_currentTriple.predicate, m_currentTriple.object };

    // Check index order
    for (int i = 0; i < 3; ++i) {
        // If query item is empty: we may have to skip some records to find a match, but there may be one
        // in the future
        if (queryStrings[i].isEmpty()) {
            return true;
        }
        // If query is not empty, and not a match anymore, we can stop
        else if (queryStrings[i] != currentStrings[i]) {
            return false;
        }
        // Else query is not empty, but matches: look into next index item
    }
    // All three index items match: continue search
    return true;
}

void BdbTripleTBtreeCursor::advanceCursor()
{
    DBT data;
    memset(&data, 0, sizeof(DBT));

    int ret = m_cursor->c_get(m_cursor, &m_current, &data, DB_NEXT);
    if (ret != 0) {
        m_eof = true;
        // Clear triple
        m_currentTriple.subject = QString();
        m_currentTriple.predicate = QString();
        m_currentTriple.object = QString();
        m_currentTriple.distance = 0;
        return;
    }
    m_currentTriple = RdfsBdbTripleT::btreeTripleFromDbt(&m_current, &data, m_primary);
}
