/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "findqueries.h"

#include <QFileInfo>
#include <QTextStream>
#include <QSet>
#include <stdlib.h>

#include "../triple.h"
#include "../rdfscursor.h"
#include "../rdfsupdate.h"
#include "../rdfslogger.h"
#include "rdfsclosurecheck.h"


using namespace Rdfs;

ChainDefinition::ChainDefinition(const QString &predicate)
{
    start = QString();
    this->predicate = predicate;
    end = QString();
    size = 0;
}

ChainDefinition::ChainDefinition(const Rdfs::Triple& t)
{
    start = t.subject;
    predicate = t.predicate;
    end = t.object;
    size = 1;
}

bool ChainDefinition::operator<(const Rdfs::ChainDefinition& b) const
{
    if (this->predicate != b.predicate) {
        return (this->predicate < b.predicate);
    }

    if (this->size != b.size) {
        return (this->size < b.size);
    }

    if (this->start != b.start) {
        return (this->start < b.start);
    }

    return (this->end < b.end);
}

ChainDefinition& ChainDefinition::operator+=(const Rdfs::ChainDefinition& b)
{
    if (b.predicate != this->predicate) {
        return *this;
    }

    // Append chain
    if (b.start == this->end) {
        this->end = b.end;
        this->size += b.size;
    }
    // Or prepend
    else if (b.end == this->start) {
        this->start = b.start;
        this->size += b.size;
    }
    return *this;
}

const ChainDefinition ChainDefinition::operator+(const Rdfs::ChainDefinition& b) const
{
    ChainDefinition c = *this;
    c += b;
    return c;
}


bool FindQueries::exec(const QStringList& arguments)
{
    const RdfsTable table = RDFS_ORIGINAL;

    int queryClass, n, valid;
    QString file;
    bool truncate, ok;

    queryClass = arguments.at(0).toInt(&ok);
    n = arguments.at(1).toInt(&ok);
    valid = arguments.at(2).toInt(&ok);

    file = (arguments.size() > 3) ? arguments.at(3) : "-";
    truncate = (arguments.size() > 4) ? (arguments.at(4) == "1") : false;

    QList<Triple> queries;

    switch(queryClass)
    {
        case 1:
        default:
            queries = generateClass1Queries(n, valid, table);
            break;
        case 2:
            queries = generateClass2Queries(n, valid, table);
            break;
        case 3:
            queries = generateClass3Queries(n, valid, table);
            break;
        case 4:
            queries = generateClass4Queries(n, valid, table);
            break;
    }

    if (queries.isEmpty()) {
        RdfsLogger::instance()->error(tr("No queries generated - aborting."));
        return false;
    }

    // Open file
    QFile f;
    if (file != "-") {
        f.setFileName(file);
        QIODevice::OpenMode flags = QIODevice::WriteOnly | QIODevice::Text;
        if (truncate) {
            flags |= QIODevice::Truncate;
        }
        else {
            flags |= QIODevice::Append;
        }
        f.open(flags);
    }
    else {
        // Open stdout -> bypass RdfsLogger class preferrably
        f.open(stdout, QIODevice::WriteOnly | QIODevice::Text);
    }
    QTextStream t(&f);

    foreach(Triple tt, queries) {
        t << tt.subject << " " << tt.predicate << " " << tt.object << endl;
    }
    f.close();
    return true;
}

QList< Triple > FindQueries::generateClass1Queries(int n, int valid, RdfsTable table) const
{
    // Class 1: Property Domain Class / Property Range Class
    Triple search;
    search.subject = QString();
    search.predicate = "dom";
    search.object = QString();
    QList<Triple> triples = framework()->store()->findTriples(search, table);
    search.predicate = "range";
    triples += framework()->store()->findTriples(search, table);

    if (triples.isEmpty()) {
        RdfsLogger::instance()->error(tr("No valid domain/range triples found."));
        return QList<Triple>();
    }

    if (triples.size() < valid) {
        RdfsLogger::instance()->log(tr("Warning: could not find enough triples to sample valid items. Less valid queries will be returned!"));
    }

    // Sample 'valid' # of triples from the data set
    QList<Triple> queries;
    while (queries.size() < valid && !triples.isEmpty()) {
        int index = qrand() % triples.size();
        queries.append(triples.takeAt(index));
    }

    // Collect properties/classes
    QList<QString> props, classes;
    search.subject = QString();
    search.predicate = "type";
    search.object = "Property";
    RdfsCursor *c = framework()->store()->cursor(search);
    if (c->open()) {
        while (!c->eof()) {
            Triple t = c->current();
            props.append(t.subject);
            c->next();
        }
        c->close();
    }
    delete c;
    search.subject = QString();
    search.predicate = "type";
    search.object = "Class";
    c = framework()->store()->cursor(search);
    if (c->open()) {
        while (!c->eof()) {
            Triple t = c->current();
            classes.append(t.subject);
            c->next();
        }
        c->close();
    }
    delete c;

    if (props.isEmpty() || classes.isEmpty()) {
        RdfsLogger::instance()->error(tr("No valid atoms of type Property or Class found."));
        return QList<Triple>();
    }

    // Now add invalid queries
    while (queries.size() < n) {
        Triple t;
        t.subject = props.at(qrand() % props.size());
        t.predicate = (chance(0.5)) ? "dom" : "range";
        t.object = classes.at(qrand() % classes.size());

        // Verify that triple doesn't exist - if it does, just start over
        if (!framework()->store()->hasTriple(t)) {
            queries.append(t);
        }
    }

    // All done :)
    return queries;
}

QList< Triple > FindQueries::generateClass2Queries(int n, int valid, Rdfs::RdfsTable table) const
{
    // Class 2: Atom Subproperty Atom - SP chain at most length 1 (i.e. one statement)
    QList<Triple> result;

    // Gather all max. chains of length 1 (P sp Q)
    // Note that we want max chains (from beginning to end), i.e. loose edges. These shouldn't be "surrounded"
    // by other sp statements
    RdfsLogger::instance()->debug("Finding disconnected chains");
    QList<ChainDefinition> chains = disconnectedChains("sp", table);
    QList<ChainDefinition> allChains = chains;
    if (chains.size() <= 1) {
        RdfsLogger::instance()->error(tr("Not enough valid sp triples found."));
        return QList<Triple>();
    }
    if (chains.size() < valid) {
        RdfsLogger::instance()->log(tr("Warning: could not find enough sp chains of length 1 to sample valid items. Less valid queries will be returned!"));
    }

    // Grab random selection of valid queries
    RdfsLogger::instance()->debug("Valid triples...");
    while (result.size() < valid && !chains.isEmpty()) {
        int index = qrand() % chains.size();
        ChainDefinition c = chains.takeAt(index);

        Triple search;
        search.subject = QString();
        search.predicate = c.start;
        search.object = QString();
        QList<Triple> instances = framework()->store()->findTriples(search, table);
        if (!instances.isEmpty()) {
            Triple i = instances.at(qrand() % instances.size());
            i.predicate = c.end;
            result.append(i);
        }
    }

    if (result.size() >= n) {
        return result;
    }

    // Invalid queries: collect roots
    RdfsLogger::instance()->debug("Invalid triples - collect roots/tails...");
    QSet<QString> roots, tails;
    foreach(ChainDefinition c, allChains) {
        roots.insert(c.start);
        tails.insert(c.end);
    }

    // Create queries
    QList<QString> rootList, tailList; // Necessary for random picking using .at()
    rootList = roots.toList();
    tailList = tails.toList();
    RdfsLogger::instance()->debug("Invalid triples - build...");
    while (result.size() < n) {
        Triple spquery;
        spquery.subject = rootList.at(qrand() % rootList.size());
        spquery.predicate = "sp";
        spquery.object = tailList.at(qrand() % tailList.size());

        // Check that chain is invalid - note that the root is constructed from
        // a disconnected sp triple, so we can just check for the existence of the triple
        if (framework()->store()->hasTriple(spquery, table)) {
            continue;
        }

        // Find valid instance of root property
        Triple iquery;
        iquery.subject = QString();
        iquery.predicate = spquery.subject;
        iquery.object = QString();

        QList<Triple> instances = framework()->store()->findTriples(iquery, table);
        // Find instance that does not exist
        while (!instances.isEmpty()) {
            Triple instance = instances.takeAt(qrand() % instances.size());
            instance.predicate = spquery.object;
            if (!framework()->store()->hasTriple(instance, table)) {
                result.append(instance);
                break;
            }
        }
    }

    return result;
}

QList< Triple > FindQueries::generateClass3Queries(int n, int valid, Rdfs::RdfsTable table)
{
    // Class 3: Atom Type Class - long inference chains

    // Load all longest SP/SC chains
    QList<ChainDefinition> spChains = collectChains("sp", table, n * 100);
    QList<Triple> validQueries, invalidQueries;

    RdfsClosureCheck *checkCommand = new RdfsClosureCheck(framework());

    // List is sorted by chain size, descending -> longest first
    RdfsLogger::instance()->debug(tr("Found %1 sp Chains").arg(spChains.size()));
    foreach(ChainDefinition c, spChains) {
        Triple search, domain, range;
        QList<Triple> searchList;

        // Locate domain/range classes
        search.subject = c.end;
        search.predicate = "dom";
        search.object = QString();
        searchList = framework()->store()->findTriples(search, table);
        if (searchList.isEmpty()) {
            RdfsLogger::instance()->debug("Dom list empty - skipping");
            continue;
        }
        domain = searchList.first();

        search.subject = c.end;
        search.predicate = "range";
        search.object = QString();
        searchList = framework()->store()->findTriples(search, table);
        if (searchList.isEmpty()) {
            RdfsLogger::instance()->debug("Range list empty - skipping");
            continue;
        }
        range = searchList.first();

        // Find longest subclass chains for domain and range
        m_maxChains = 0;

        QList<ChainDefinition> domainChains, rangeChains;
        m_traversedChains.clear();
        traverse("sc", table, domain.object);
        RdfsLogger::instance()->debug("Finished traverse for domain");
        domainChains = m_traversedChains;

        m_traversedChains.clear();
        traverse("sc", table, range.object);
        RdfsLogger::instance()->debug("Finished traverse for range");
        rangeChains = m_traversedChains;

        if (domainChains.isEmpty() && rangeChains.isEmpty()) {
            RdfsLogger::instance()->debug("Dom & range chains empty - skipping");
            continue;
        }

        if (!domainChains.isEmpty()) {
            RdfsLogger::instance()->debug("Dom not empty");
            qSort(domainChains);
        }
        if (!rangeChains.isEmpty()) {
            RdfsLogger::instance()->debug("Range not empty");
            qSort(rangeChains);
        }

        // domainChains, rangeChains are now sorted by length
        // Find the longest possible chain
        bool useDomain = (!domainChains.isEmpty() && !rangeChains.isEmpty())
                       ? (rangeChains.last() < domainChains.last()) // Use domain if chain is longer than range
                       : !domainChains.isEmpty(); // If either domain/range is empty, use domain if not empty
        ChainDefinition scChain = (useDomain) ? domainChains.last() : rangeChains.last();

        // Find any instance of the starting property
        Triple iSearch;
        iSearch.subject = QString();
        iSearch.predicate = c.start;
        iSearch.object = QString();
        QList<Triple> instances = framework()->store()->findTriples(iSearch, table);

        if (instances.isEmpty()) {
            continue; // No instances
        }

        // Build final query
        Triple query;
        if (validQueries.size() < valid) {
            Triple instance = instances.at(qrand() % instances.size());
            query.subject = (useDomain) ? instance.subject : instance.object;
            query.predicate = "type";
            query.object = scChain.end;

            validQueries.append(query);
        }
        if (invalidQueries.size() < n - valid && (domainChains.isEmpty() || rangeChains.isEmpty() || domainChains.last().end != rangeChains.last().end)) {
            // To make invalid query: check for type of wrong item (e.g. for domain, use the range of the instance,
            // for range, use the domain of the instance).

            // To ensure query is really invalid, check that it does not exist.
            Triple instance = instances.at(qrand() % instances.size());
            query.subject = (useDomain) ? instance.object : instance.subject;
            query.predicate = "type";
            query.object = scChain.end;

            if (!checkCommand->checkClosure(query, table)) {
                invalidQueries.append(query);
            }
        }

        if (validQueries.size() + invalidQueries.size() >= n) {
            break;
        }
    }

    delete checkCommand;

    return validQueries + invalidQueries;
}

QList< Triple > FindQueries::generateClass4Queries(int n, int valid, Rdfs::RdfsTable table)
{
    // Class 3: Atom Type Class - long inference chains

    // Load all longest SP/SC chains
    QList<ChainDefinition> spChains = collectChains("sp", table, n*10);
    QList<ChainDefinition> scChains = collectChains("sc", table, n*10);
    QList<Triple> validQueries, invalidQueries;
    RdfsClosureCheck *checkCommand = new RdfsClosureCheck(framework());

    // List is sorted by chain size, descending -> longest first
    RdfsLogger::instance()->debug(tr("Found %1 sp Chains").arg(spChains.size()));
    QList<ChainDefinition>::const_iterator i, j;
    i = spChains.constBegin();
    j = scChains.constBegin();
    while (validQueries.size() + invalidQueries.size() < n && !(i == spChains.constEnd() && j == scChains.constEnd())) {
        ChainDefinition c, next;
        QList<ChainDefinition> otherChains;
        if (i != spChains.constEnd() && (j == scChains.constEnd() || (*i).size >= (*j).size)) {
            c = *i;
            ++i;
            otherChains = spChains;
        }
        else {
            c = *j;
            ++j;
            otherChains = scChains;
        }

        // Add valid query
        Triple query;
        query.subject = c.start;
        query.predicate = c.predicate;
        query.object = c.end;

        if (validQueries.size() < valid && !validQueries.contains(query)) {
            validQueries.append(query);
        }

        if (invalidQueries.size() < n - valid) {
            // Pick a random chain from the list
            next = otherChains.at(qrand() % otherChains.size());

            // Choose whether to change domain or range
            if (chance(0.5)) {
                query.subject = next.start;
            }
            else {
                query.object = next.end;
            }
            // Make sure it really doesn't exist
            if (!invalidQueries.contains(query) && !checkCommand->checkClosure(query, table)) {
                invalidQueries.append(query);
            }
        }
    }
    delete checkCommand;

    return validQueries + invalidQueries;
}

QList< ChainDefinition > FindQueries::disconnectedChains(const QString& predicate, RdfsTable table) const
{
    Triple search;
    search.subject = QString();
    search.predicate = predicate;
    search.object = QString();

    RdfsCursor *c = framework()->store()->cursor(search, table);
    if (!c->open()) {
        delete c;
        return QList<ChainDefinition>();
    }

    QList< ChainDefinition > result;
    while (!c->eof()) {
        Triple t = c->current();
        Triple s1, s2, s3, s4;
        // May not exist
        s1.subject = QString();
        s1.predicate = t.predicate;
        s1.object = t.subject;

        // May match only current triple
        s2.subject = QString();
        s2.predicate = t.predicate;
        s2.object = t.object;

        // May match only current triple
        s3.subject = t.subject;
        s3.predicate = t.predicate;
        s3.object = QString();

        // May not exist
        s4.subject = t.object;
        s4.predicate = t.predicate;
        s4.object = QString();

        if (!framework()->store()->hasTriple(s1, table) && framework()->store()->checkTriple(s2, table) == 1  &&
            framework()->store()->checkTriple(s3, table) == 1 && !framework()->store()->hasTriple(s4, table)) {
            RdfsLogger::instance()->debug(QString("Found one: %1 %2 %3").arg(t.subject, t.predicate, t.object));
            result.append(t);
        }

        c->next();
    }
    c->close();
    delete c;

    return result;
}

QList< ChainDefinition > FindQueries::collectChains(const QString& predicate, RdfsTable table, int maxAmount)
{
    m_traversedChains.clear();
    m_maxChains = maxAmount;
    traverse(predicate, table);

    return m_traversedChains;
}

void FindQueries::foundChain(const QList< QString >& atoms, const QString& predicate)
{
    ChainDefinition c(predicate);
    c.start = atoms.first();
    c.end = atoms.last();
    c.size = atoms.size() - 1;

    bool append = (m_traversedChains.isEmpty()) ? true : c < m_traversedChains.last();

    if (append && m_maxChains != 0 && m_traversedChains.size() >= m_maxChains) {
        RdfsLogger::instance()->debug("Not appending");
        return;
    }
    else if (append) {
        m_traversedChains.append(c);
        RdfsLogger::instance()->debug("Appending");
        return;
    }

    // Binary search for proper position
    int i = 0;
    int j = m_traversedChains.size();
    while (i < j) {
        int m = (i + j) / 2;

        if (c < m_traversedChains.at(m)) {
            i = m + 1; // Move towards end of list
        }
        else {
            j = m; // Move towards start of list
        }
    }
    RdfsLogger::instance()->debug(QString("Inserting at %1").arg(i));
    m_traversedChains.insert(i, c);
    if (m_maxChains > 0 && m_traversedChains.size() > m_maxChains) {
        m_traversedChains.removeLast();
    }
}

void FindQueries::skipTriple(const Rdfs::Triple& t)
{
    RdfsLogger::instance()->debug("Skipping...");
}

void FindQueries::startChain(const Rdfs::Triple& t)
{
    RdfsLogger::instance()->debug("New chain...");
}


bool FindQueries::chance(float p) const
{
    int rgn = qrand();
    return (rgn < qRound(p * (float)RAND_MAX));
}

QString FindQueries::checkArguments(const QStringList& arguments) const
{
    int n, valid;
    QString file;
    bool truncate, ok;

    if (arguments.at(0) != "1" && arguments.at(0) != "2" && arguments.at(0) != "3" && arguments.at(0) != "4") {
        return tr("Unknown query class.");
    }

    n = arguments.at(1).toInt(&ok);
    if (!ok || n <= 0) {
        return tr("No queries to generate.");
    }
    valid = arguments.at(2).toInt(&ok);
    if (!ok || valid > n) {
        return tr("Will not generate more valid queries than the total amount of queries.");
    }

    if (arguments.size() > 3) {
        // File argument given
        file = arguments.at(3);
        if (file != "-") {
            // Check if file is writable
            QFileInfo f(file);
            QFileInfo d(f.path());

            if (f.isDir()) {
                return tr("File %1 is a directory.").arg(file);
            }
            else if (f.exists() && !f.isWritable()) {
                return tr("File %1 is not writable.").arg(file);
            }
            else if (!f.exists() && (!d.exists() || !d.isDir() || !d.isWritable())) {
                return tr("Cannot create file %1.").arg(file);
            }
        }

        if (arguments.size() > 4) {
            if (arguments.at(4) != "0" && arguments.at(4) != "1") {
                return tr("Truncate argument must be 0 or 1.");
            }
            truncate = (arguments.at(4) == "1");
        }
        else {
            truncate = 0;
        }

        if (file == "-" && truncate) {
            return tr("Cannot truncate stdout.");
        }
    }
    return QString();

}

QList< Rdfs::CommandParameter > FindQueries::arguments() const
{
    CommandParameter queryClass, n, valid, file, truncate;
    queryClass.name = tr("class");
    queryClass.help = tr("The class (1, 2, 3) of queries to generate.\n"
                         "Class 1: A simple lookup (Property Domain Class).\n"
                         "Class 2: A single traversal step (Atom Subproperty Atom).\n"
                         "Class 3: A long query (Atom Type Class).\n");
    queryClass.examples << "1";
    queryClass.required = true;

    n.name = tr("n");
    n.help = tr("The total amount of queries to generate (valid and invalid).");
    n.examples << "100";
    n.required = true;

    valid.name = tr("valid");
    valid.help = tr("The amount of valid queries to generate (queries which exist in the closure).\n");
    valid.examples << "50";
    valid.required = true;

    file.name = tr("file");
    file.help = tr("The file to write the queries to. Use - for stdout (default).");
    file.examples << "-" << tr("queries.rdfsupdate");
    file.required = false;

    truncate.name = tr("truncate");
    truncate.help = tr("Whether to truncate the file to write queries to (0 or 1).\nCannot be used if file is stdout (-).\n"
                       "If 0, the queries will be appended to the file. Default is 0.");
    truncate.examples << "0";
    truncate.required = false;

    return QList<CommandParameter>() << queryClass << n << valid << file << truncate;

}

FindQueries::FindQueries(Rdfs::RdfsUpdate* parent): RdfsChainCommand(parent)
{
    m_maxChains = 0;
}

FindQueries::~FindQueries()
{

}

#include "findqueries.moc"
