/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "instancegenerate.h"

#include <QList>
#include <QMap>
#include <stdlib.h>

#include "../rdfslogger.h"
#include "../rdfsupdate.h"
#include "../rdfscursor.h"

using namespace Rdfs;

InstanceGenerate::InstanceGenerate(RdfsUpdate* parent)
  : RdfsCommand(parent)
{
}

bool InstanceGenerate::exec(const QStringList& arguments)
{
    // Will succeed, as checkArguments() has passed
    RdfsTable table = RDFS_ORIGINAL;

    bool ok;
    int m = arguments.at(0).toInt(&ok);
    QString atomPrefix = (arguments.size() > 1) ? arguments.at(1) : "Atom";
    QString classType = (arguments.size() > 2) ? arguments.at(2) : "Class";

    // Find classes
    Triple csearch;
    csearch.subject = QString();
    csearch.predicate = "type";
    csearch.object = classType;

    QStringList classes;
    QMap<QString, int> classMap; // For faster lookups

    RdfsLogger::instance()->log(tr("Loading classes..."));
    RdfsCursor *c = framework()->store()->cursor(csearch, table);
    if (!c->open()) {
        RdfsLogger::instance()->error(tr("No classes (type %1) found!").arg(classType));
        delete c;
        return false;
    }

    while (!c->eof()) {
        classes << c->current().subject;
        classMap.insert(c->current().subject, classes.size() - 1);
        c->next();
    }
    c->close();
    delete c;

    framework()->store()->startBulkLoad();

    int n = classes.size();
    for (int i = 0; i < n; ++i) {
        QString className = classes.at(i);
        RdfsLogger::instance()->log(tr("Generating instances for class %1").arg(className));

        // Get list of possible properties
        QList<Triple> domainList, rangeList, resultList, propList;
        Triple t;
        t.subject = QString();
        t.object = className;

        t.predicate = "dom";
        domainList = framework()->store()->findTriples(t, table);

        t.predicate = "range";
        rangeList = framework()->store()->findTriples(t, table);

        // For each domain/range list item, find resp. range/domain of property
        // and add (domain, property, range) to propList
        Q_FOREACH(t, domainList) {
            Triple tt;
            tt.subject = t.subject;
            tt.predicate = "range";
            tt.object = QString();

            resultList = framework()->store()->findTriples(tt, table);
            if (!resultList.isEmpty() && classMap.contains(resultList.at(0).object)) {
                tt.subject = t.object;
                tt.predicate = t.subject;
                tt.object = resultList.at(0).object;
                propList << tt;
            }
        }

        Q_FOREACH(t, rangeList) {
            Triple tt;
            tt.subject = t.subject;
            tt.predicate = "domain";
            tt.object = QString();

            resultList = framework()->store()->findTriples(tt, table);
            if (!resultList.isEmpty() && classMap.contains(resultList.at(0).object)) {
                tt.subject = resultList.at(0).object;
                tt.predicate = t.subject;
                tt.object = t.object;
                propList << tt;
            }
        }


        for (int j = 0; j < m; ++j) {
            QString atomName = atomPrefix + QString::number(i*m + j);
            Triple t;

            if (propList.isEmpty() || chance(0.25)) {
                t.subject = atomName;
                t.predicate = "type";
                t.object = className;
                framework()->store()->insertTriple(t);
            }
            else {
                int index = qrand() % propList.count();
                Triple prop = propList.at(index);
                bool useDomain = (prop.subject == className && prop.object == className)
                               ? chance(0.5)
                               : (prop.subject == className);

                if (useDomain) {
                    // Use atom as domain (subject)
                    int c2num = classMap.value(prop.object);
                    QString a2 = atomPrefix + QString::number(c2num*m + (rand() % m));
                    t.subject = atomName;
                    t.predicate = prop.predicate;
                    t.object = a2;
                    framework()->store()->insertTriple(t);
                }
                else {
                    // Use atom as range (object)
                    int c2num = classMap.value(prop.subject);
                    QString a2 = atomPrefix + QString::number(c2num*m + (rand() % m));
                    t.subject = a2;
                    t.predicate = prop.predicate;
                    t.object = atomName;
                    framework()->store()->insertTriple(t);
                }
            }
        }
    }
    framework()->store()->endBulkLoad();

    return true;
}

bool InstanceGenerate::chance(float p)
{
    int rgn = qrand();
    return (rgn < qRound(p * (float)RAND_MAX));
}

QString InstanceGenerate::checkArguments(const QStringList& arguments) const
{
    bool ok;
    int m = arguments.at(0).toInt(&ok);
    if (!ok) {
        return tr("Given amount of instances per class (m) is not an integer");
    }
    if (m <= 0) {
        return tr("Amount of instances per class (m) must be at least 1");
    }
    return QString();
}

bool InstanceGenerate::requireConnection() const
{
    return true;
}

QList< Rdfs::CommandParameter > InstanceGenerate::arguments() const
{
    QList<Rdfs::CommandParameter> list;
    Rdfs::CommandParameter m;
    m.name = "m";
    m.examples << "10";
    m.help = tr("Amount of instances to generate, per class");
    m.required = true;
    list.append(m);


    Rdfs::CommandParameter atomName;
    atomName.name = "atomName";
    atomName.examples << "Atom";
    atomName.help = tr("The base name (prefix) to use for created atoms.");
    atomName.required = false;
    list.append(atomName);

    Rdfs::CommandParameter classType;
    classType.name = "classtype";
    classType.examples << "Class";
    classType.help = tr("The type used to identify classes. Triples \"X type classtype\" must exist to indicate that X is a class.");
    classType.required = false;
    list.append(classType);

    return list;
}

QStringList InstanceGenerate::aliases() const
{
    return QStringList() << "gi";
}

QString InstanceGenerate::description() const
{
    return tr("Generate instance data for an ontology");
}

QString InstanceGenerate::name() const
{
    return "instantiate";
}

InstanceGenerate::~InstanceGenerate()
{

}

#include "instancegenerate.moc"
