/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "rdfsaddrule.h"

#include <QRegExp>
#include <QMap>

#include "../rdfsupdate.h"
#include "../triple.h"
#include "../rdfssql.h"

using namespace Rdfs;


RdfsAddRule::RdfsAddRule(RdfsUpdate* parent): RdfsCommand(parent) {
}


RdfsAddRule::~RdfsAddRule() {
}


bool RdfsAddRule::exec(const QStringList& arguments) {
    Triple t1, t2;
    t1.subject = arguments.at(0);
    t1.predicate = arguments.at(1);
    t1.object = arguments.at(2);

    t2.subject = arguments.at(3);
    t2.predicate = arguments.at(4);
    t2.object = arguments.at(5);

    // Insert rule
    framework()->store()->insertRule(t1, t2);
    return true;
}

QString RdfsAddRule::checkArguments(const QStringList& arguments) const {
    // Check that variables on the left occur on the right too
    const QRegExp r("@[0-9]+");
    QStringList lhs = QStringList() << arguments.at(0) << arguments.at(1) << arguments.at(2);

    for (int i = 3; i < 6; ++i) {
        QString item = arguments.at(i);
        if (r.exactMatch(item) && !lhs.contains(item)) {
            return tr("Invalid rule: variable \"%1\" doesn't occur in left-hand side triple.").arg(item);
        }
    }

    // Check that no values are blanks
    const QRegExp globalVarRegex("\\$[0-9]+");
    for (int i = 0; i < 6; ++i) {
        if (arguments.at(i) == "_") {
            return tr("Blank nodes are not supported.");
        }
        else if (globalVarRegex.exactMatch(arguments.at(i))) {
            return tr("Syntax '$[0-9]+' is reserved for internal use.");
        }
    }

    // All is OK
    return QString();

}

bool RdfsAddRule::requireConnection() const {
    return true;
}

QList< Rdfs::CommandParameter > RdfsAddRule::arguments() const {
    CommandParameter s1;
    s1.name = tr("subject1");
    s1.help = tr("The subject of the left-hand side triple. Variables are denoted as \"@n\" where n is a positive integer number");
    s1.examples = QStringList() << "@1" << "abc";
    s1.required = true;

    CommandParameter p1;
    p1.name = tr("predicate1");
    p1.help = tr("The predicate of the left-hand side triple. Variables are denoted as \"@n\" where n is a positive integer number");
    p1.examples = QStringList() << "@1" << "abc";
    p1.required = true;

    CommandParameter o1;
    o1.name = tr("object1");
    o1.help = tr("The object of the left-hand side triple. Variables are denoted as \"@n\" where n is a positive integer number");
    o1.examples = QStringList() << "@1" << "abc";
    o1.required = true;


    CommandParameter s2;
    s2.name = tr("subject2");
    s2.help = tr("The subject of the right-hand side triple. Variables are denoted as \"@n\" where n is a positive integer number");
    s2.examples = QStringList() << "@1" << "abc";
    s2.required = true;

    CommandParameter p2;
    p2.name = tr("predicate2");
    p2.help = tr("The predicate of the right-hand side triple. Variables are denoted as \"@n\" where n is a positive integer number");
    p2.examples = QStringList() << "@1" << "abc";
    p2.required = true;

    CommandParameter o2;
    o2.name = tr("object2");
    o2.help = tr("The object of the right-hand side triple. Variables are denoted as \"@n\" where n is a positive integer number");
    o2.examples = QStringList() << "@1" << "abc";
    o2.required = true;

    return (QList<CommandParameter>() << s1 << p1 << o1 << s2 << p2 << o2);
}

QStringList RdfsAddRule::aliases() const {
    return QStringList() << "ar";
}

QString RdfsAddRule::description() const {
    return tr("Add a new deterministic tuple generating dependency. Note that all variables in the right-hand side triple must also occur on the left-hand side.");
}

QString RdfsAddRule::name() const {
    return "addRule";
}

#include "rdfsaddrule.moc"
