/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "rdfschainstats.h"

#include "../rdfsupdate.h"
#include "../rdfscursor.h"
#include "../rdfslogger.h"

using namespace Rdfs;

RdfsChainStats::RdfsChainStats(RdfsUpdate* parent): RdfsChainCommand(parent)
{

}

QList< Rdfs::CommandParameter > RdfsChainStats::arguments() const
{
    Rdfs::CommandParameter property;
    property.name = tr("property");
    property.required = true;
    property.help = tr("The transitive property to retrieve statistics from.");
    property.examples = QStringList() << "sc" << "sp";

    Rdfs::CommandParameter table;
    table.name = tr("table");
    table.required = false;
    table.help = tr("The graph name (original, update) to get the statistics from.");
    table.examples = QStringList() << "original" << "update";

    return QList<CommandParameter>() << property << table ;
}

QString  RdfsChainStats::checkArguments(const QStringList& arguments) const
{
    if (arguments.at(0).isEmpty()) {
        return tr("Empty property given.");
    }

    if (arguments.count() == 2 && arguments.at(1).compare("original", Qt::CaseInsensitive) != 0 &&
        arguments.at(1).compare("update", Qt::CaseInsensitive) != 0) {
            return tr("Unknown table %1 given.").arg(arguments.at(1));
    }

    return QString();
}

bool RdfsChainStats::exec(const QStringList& arguments)
{
    QString predicate = arguments.at(0);
    RdfsTable table = RDFS_ORIGINAL;

    if (arguments.count() > 1 && arguments.at(1).compare("original", Qt::CaseInsensitive) == 0) {
        table = RDFS_ORIGINAL;
    }
    else if (arguments.count() > 1 && arguments.at(1).compare("update", Qt::CaseInsensitive) == 0) {
        table = RDFS_UPDATE;
    }

    // Compute statistics
    resetStats();
    // This will call the necessary methods to compute stats:
    traverse(predicate, table);

    // Post-processing: root/leaves list
    QStringList rootList = m_roots.toList();
    QStringList leavesList = m_leaves.toList();
    qSort(rootList);
    qSort(leavesList);

    // Keep top-10 degrees
    QList<AtomDegree> topInDegree = top(m_inDegree, 10);
    QList<AtomDegree> topOutDegree = top(m_outDegree, 10);
    QList<AtomDegree> topTotalDegree = top(m_totalDegree, 10);
    QStringList topInDegreeStrings, topOutDegreeStrings, topTotalDegreeStrings;

    // Print stats
    RdfsLogger::instance()->log("============================");
    RdfsLogger::instance()->log(tr("Statistics for transitive predicate: %1").arg(predicate));
    RdfsLogger::instance()->log(tr("- Total amount of statements: %1").arg(m_tripleCount));
    RdfsLogger::instance()->log(tr("- Amount of unique chains: %1").arg(m_stats.count));
    RdfsLogger::instance()->log(tr("- Minimum chain length: %1").arg(m_stats.minChain.size()));
    RdfsLogger::instance()->log(tr("- Maximum chain length: %1").arg(m_stats.maxChain.size()));
    RdfsLogger::instance()->log(tr("- Average chain length: %1").arg(m_stats.avgLength, 0, 'f', 3));
    RdfsLogger::instance()->log("============================");
    RdfsLogger::instance()->log(tr("Roots:"));
    RdfsLogger::instance()->log(rootList.join(", "));
    RdfsLogger::instance()->log("============================");
    RdfsLogger::instance()->log(tr("Leaves:"));
    RdfsLogger::instance()->log(leavesList.join(", "));
    RdfsLogger::instance()->log("============================");
    RdfsLogger::instance()->log(tr("Minimal chain (%1):").arg(QString::number(m_stats.minChain.size())));
    RdfsLogger::instance()->log(m_stats.minChain.join(" " + predicate + " "));
    RdfsLogger::instance()->log(tr("Maximal chain (%1):").arg(QString::number(m_stats.maxChain.size())));
    RdfsLogger::instance()->log(m_stats.maxChain.join(" " + predicate + " "));
    RdfsLogger::instance()->log("============================");
    RdfsLogger::instance()->log(tr("Top 10 in-degree atoms:"));
    Q_FOREACH(AtomDegree d, topInDegree) {
        topInDegreeStrings << tr("%1 (%2)").arg(d.first).arg(d.second);
    }
    RdfsLogger::instance()->log(topInDegreeStrings.join(", "));
    RdfsLogger::instance()->log("============================");
    RdfsLogger::instance()->log(tr("Top 10 out-degree atoms:"));
    Q_FOREACH(AtomDegree d, topOutDegree) {
        topOutDegreeStrings << tr("%1 (%2)").arg(d.first).arg(d.second);
    }
    RdfsLogger::instance()->log(topOutDegreeStrings.join(", "));
    RdfsLogger::instance()->log("============================");
    RdfsLogger::instance()->log(tr("Top 10 total-degree atoms:"));
    Q_FOREACH(AtomDegree d, topTotalDegree) {
        topTotalDegreeStrings << tr("%1 (%2)").arg(d.first).arg(d.second);
    }
    RdfsLogger::instance()->log(topTotalDegreeStrings.join(", "));
    RdfsLogger::instance()->log("============================");
    return true;
}

QList<AtomDegree> RdfsChainStats::top(QMap< QString, quint64 > map, int count)
{
    QList<AtomDegree> result;

    QMap<QString, quint64>::const_iterator i;
    for (i = map.constBegin(); i != map.constEnd(); ++i) {
        AtomDegree current;
        current.first = i.key();
        current.second = i.value();
        // Find position in result set using some sort of binary search
        int j = 0;
        int k = result.size();
        while (j < k) {
            int mid = (j + k) / 2;
            if (current < result.at(mid)) {
                j = mid + 1;
            }
            else {
                k = mid;
            }
        }
        // Note: j == k. Insert current at index j. Then pop last item.
        // Note that behaviour of QList::insert() at index size() means the
        // item is appended. As such it may be added and removed, which is ok.
        result.insert(j, current);
        if (result.size() > count) {
            result.removeLast();
        }
    }

    return result;
}

void RdfsChainStats::resetStats()
{
    m_tripleCount = 0;
    m_stats.count = 0;
    m_stats.avgLength = 0.0;
    m_stats.minChain.clear();
    m_stats.maxChain.clear();

    m_inDegree.clear();
    m_outDegree.clear();
    m_totalDegree.clear();

    m_roots.clear();
    m_leaves.clear();
}

void RdfsChainStats::foundChain(const QList< QString >& atoms, const QString& predicate)
{
    RdfsLogger::instance()->debug(QString("Found chain length %1: %2").arg(atoms.size()).arg(QStringList(atoms).join(" " + predicate + " ")));
    // Length stats
    // Note: 'atoms.size() - 1' is the amount of triples in the chain (atoms.size() >= 2).
    m_stats.avgLength = ((m_stats.avgLength * m_stats.count) + (atoms.size() - 1)) / (float)(m_stats.count + 1);
    ++m_stats.count;

    // Min/max chain
    if (atoms.size() > m_stats.maxChain.size()) {
        m_stats.maxChain = atoms;

    }
    if (atoms.size() < m_stats.minChain.size() || m_stats.minChain.isEmpty()) {
        m_stats.minChain = atoms;
    }

    // Add roots/leaves
    m_roots << atoms.first();
    m_leaves << atoms.last();
}

void RdfsChainStats::startChain(const Rdfs::Triple& t)
{
    RdfsLogger::instance()->debug(QString("New start (%1 %2 %3)").arg(t.subject, t.predicate, t.object));
    countTriple(t);
}

void RdfsChainStats::skipTriple(const Rdfs::Triple& t)
{
    RdfsLogger::instance()->debug(QString("Skipping (%1 %2 %3)").arg(t.subject, t.predicate, t.object));
    countTriple(t);
}

void RdfsChainStats::countTriple(const Rdfs::Triple& t)
{
    ++m_tripleCount;
    // Add in/out/totaldegree stats
    incDegree(&m_inDegree, t.object);
    incDegree(&m_outDegree, t.subject);
    incDegree(&m_totalDegree, t.subject);
    incDegree(&m_totalDegree, t.object);
}

void RdfsChainStats::incDegree(QMap< QString, quint64 >* degreeList, const QString& atom)
{
    if (degreeList->contains(atom)) {
        QMap<QString, quint64>::iterator i = degreeList->find(atom);
        i.value() = i.value() + 1;
    }
    else {
        degreeList->insert(atom, 1);
    }
}

RdfsChainStats::~RdfsChainStats()
{

}

#include "rdfschainstats.moc"
