/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "rdfsclosurecheck.h"

#include <QQueue>
#include <QSet>
#include <sys/resource.h>

#include "../triple.h"
#include "../rdfsupdate.h"
#include "../rdfslogger.h"

#include "../rdfsrules/subpropertyrule.h"
#include "../rdfsrules/subclassrule.h"
#include "../rdfsrules/typerule.h"
#include "rdfsstats.h"

using namespace Rdfs;

RdfsClosureCheck::RdfsClosureCheck(RdfsUpdate* parent)
    : RdfsTripleCommand(parent)
{
}

QString RdfsClosureCheck::checkArguments(const QStringList& arguments) const {
    QString baseResult = RdfsTripleCommand::checkArguments(arguments);
    if (baseResult.isEmpty()) {
        // Check that triple does not contain empty items
        RdfsTable table;
        Triple t;
        extractArguments(arguments, &table, &t);

        if (t.subject.isEmpty() || t.predicate.isEmpty() || t.object.isEmpty()) {
            return tr("The triple contains an empty value. This is not supported.");
        }
    }
    return baseResult;
}

bool RdfsClosureCheck::exec(const QStringList& arguments) {
    // Get triple and table from arguments
    RdfsTable table;
    Triple t;
    RdfsStats *statsCommand = dynamic_cast<RdfsStats *>(framework()->command("stats"));
    extractArguments(arguments, &table, &t);

    CacheStats preStats = framework()->store()->statistics();
    if (statsCommand && (framework()->store()->capabilities() & RDFS_CACHESTATS)) {
        RdfsLogger::instance()->log("Statistics pre-algorithm:");
        statsCommand->print();
    }

    rusage start, end;
    timeval realstart, realend;

    gettimeofday(&realstart, NULL);
    getrusage(RUSAGE_SELF, &start);

    if (checkClosure(t, table)) {
        RdfsLogger::instance()->log(tr("Triple (%1, %2, %3) found in entailment!").arg(t.subject, t.predicate, t.object));
    }
    else {
        RdfsLogger::instance()->log(tr("Triple (%1, %2, %3) NOT found in entailment.").arg(t.subject, t.predicate, t.object));
    }
    getrusage(RUSAGE_SELF, &end);
    gettimeofday(&realend, NULL);

    CacheStats postStats = framework()->store()->statistics();
    if (statsCommand && (framework()->store()->capabilities() & RDFS_CACHESTATS)) {
        RdfsLogger::instance()->log("Statistics post-algorithm:");
        statsCommand->print();
        RdfsLogger::instance()->log("Statistics (difference):");
        statsCommand->print(postStats - preStats);
    }

    timeval real = timediff(realstart, realend);
    timeval user = timediff(start.ru_utime, end.ru_utime);
    timeval sys = timediff(start.ru_stime, end.ru_stime);

    RdfsLogger::instance()->log("Real time: " + QString("%1.%2").arg(real.tv_sec, 6, 10, QLatin1Char(' ')).arg(real.tv_usec, 6, 10, QLatin1Char('0')));
    RdfsLogger::instance()->log("User time: " + QString("%1.%2").arg(user.tv_sec, 6, 10, QLatin1Char(' ')).arg(user.tv_usec, 6, 10, QLatin1Char('0')));
    RdfsLogger::instance()->log("System time: " + QString("%1.%2").arg(sys.tv_sec, 5, 10, QLatin1Char(' ')).arg(sys.tv_usec, 6, 10, QLatin1Char('0')));
    RdfsLogger::instance()->log("Page faults no I/O: " + QString("%1").arg(QString::number(end.ru_minflt - start.ru_minflt)));
    RdfsLogger::instance()->log("Page faults with I/O: " + QString("%1").arg(QString::number(end.ru_majflt - start.ru_majflt)));
    RdfsLogger::instance()->log("File system input: " + QString("%1").arg(QString::number(end.ru_inblock - start.ru_inblock)));
    RdfsLogger::instance()->log("File system output: " + QString("%1").arg(QString::number(end.ru_oublock - start.ru_oublock)));
    RdfsLogger::instance()->log("====");

    return true;
}

bool RdfsClosureCheck::checkClosure(Triple t, RdfsTable table) {
    bool tInEntailment = false;

    RdfsBackend *store = framework()->store();

    // Also deals with step 1: p in {dom, range} or anything dom/range-like.
    // Hence, they are no special rules.
    if (store->checkTriple(t, table) > 0) {
        return true;
    }

    // Load rules, check them
    QList<RdfsRule *> rules = rdfsRules(table);
    Q_FOREACH(RdfsRule *r, rules) {
        if (r->check(t, table, store)) {
            tInEntailment = true;
            break;
        }
    }

    // Clean up rules
    qDeleteAll(rules.begin(), rules.end());

    // No rule matched
    return tInEntailment;
}

RdfsGraph RdfsClosureCheck::getGraph(QString predicate, RdfsTable table)
{
    Triple t;
    t.subject = "";
    t.predicate = predicate;
    t.object = "";

    RdfsGraph g;
    g.add(framework()->store()->findTriples(t, table));
    return g;
}

QList<RdfsRule *> RdfsClosureCheck::rdfsRules(RdfsTable table)
{
    QList<RdfsRule *> rules;

    // Add default rules
    RdfsRule *sp = new SubpropertyRule();
    rhodfMap(sp);
    rules << sp;

    RdfsRule *sc = new SubclassRule();
    rhodfMap(sc);
    rules << sc;

    RdfsRule *type = new TypeRule();
    rhodfMap(type);
    rules << type;
/*
    // Find "rdfsupdate:rdfsRule" triples
    RdfsGraph g = getGraph("rdfsupdate:rdfsRule", table);
    Q_FOREACH(Triple t, g.edges()) {
        // Initialize rule;
        RdfsRule *r = 0;

        if (t.object == "rdfsupdate:ruleType:sp") {
            r = new SubpropertyRule();
        }
        else if (t.object == "rdfsupdate:ruleType:sc") {
            r = new SubclassRule();
        }
        else if (t.object == "rdfsupdate:ruleType:type") {
            r = new TypeRule();
        }
        else {
            RdfsLogger::instance()->error(QString("Invalid rule type %1").arg(t.object));
            continue; // Invalid
        }

        // Load mapping for rule
        loadMap(r, t.subject, "rdfsupdate:ruleMap:sp", RDFS_SP, table);
        loadMap(r, t.subject, "rdfsupdate:ruleMap:sc", RDFS_SC, table);
        loadMap(r, t.subject, "rdfsupdate:ruleMap:type", RDFS_TYPE, table);
        loadMap(r, t.subject, "rdfsupdate:ruleMap:dom", RDFS_DOM, table);
        loadMap(r, t.subject, "rdfsupdate:ruleMap:range", RDFS_RANGE, table);

        rules << r;
    }*/

    return rules;
}

void RdfsClosureCheck::loadMap(RdfsRule* r, QString ruleName, QString mapPredicate, RhoDF index, RdfsTable table)
{
    Triple t;
    t.subject = ruleName;
    t.predicate = mapPredicate;
    t.object = QString();

    QList<Triple> mappings = framework()->store()->findTriples(t, table);
    if (!mappings.isEmpty()) {
        // Get first item
        Triple tt = mappings.at(0);
        // Assign mapping
        r->setMap(index, tt.object);
    }
}

void RdfsClosureCheck::rhodfMap(RdfsRule* r)
{
    r->setMap(RDFS_SP, "sp");
    r->setMap(RDFS_SC, "sc");
    r->setMap(RDFS_TYPE, "type");
    r->setMap(RDFS_DOM, "dom");
    r->setMap(RDFS_RANGE, "range");
}

timeval RdfsClosureCheck::timediff(const timeval& a, const timeval& b)
{
    timeval result;
    result.tv_sec = b.tv_sec - a.tv_sec;
    if (b.tv_usec >= a.tv_usec) {
        result.tv_usec = b.tv_usec - a.tv_usec;
    }
    else {
        result.tv_sec--;
        result.tv_usec = 1000000 - (a.tv_usec - b.tv_usec);
    }
    return result;
}


RdfsClosureCheck::~RdfsClosureCheck()
{
}


#include "rdfsclosurecheck.moc"
