/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "rdfsdotexport.h"

#include <QFile>
#include <QFileInfo>
#include <QTextStream>
#include <stdio.h>

#include "../triple.h"
#include "../rdfscursor.h"
#include "../rdfsupdate.h"

using namespace Rdfs;

bool RdfsDotExport::exec(const QStringList& arguments)
{
    QFile f;
    if (arguments.at(1) != "-") {
        f.setFileName(arguments.at(1));
        f.open(QIODevice::Truncate | QIODevice::WriteOnly | QIODevice::Text);
    }
    else {
        // Open stdout -> bypass RdfsLogger class preferrably
        f.open(stdout, QIODevice::WriteOnly | QIODevice::Text);
    }
    QTextStream t(&f);

    t << "digraph " << arguments.at(0) << " {" << endl;
    Triple tt;
    tt.predicate = arguments.at(0);

    Rdfs::RdfsCursor *c = framework()->store()->cursor(tt);
    if (c->open()) {
        while (!c->eof()) {
            tt = c->current();
            t << tt.subject << " -> " << tt.object << endl;
            c->next();
        }
        c->close();
    }
    t << "}";

    delete c;

    return true;
}

QString RdfsDotExport::checkArguments(const QStringList& arguments) const
{
    if (arguments.at(1) != "-") {
        // Check if file is writable
        QFileInfo file(arguments.at(1));
        QFileInfo dir(file.path());

        if (file.isDir()) {
            return tr("File %1 is a directory.").arg(arguments.at(1));
        }
        else if (file.exists() && !file.isWritable()) {
            return tr("File %1 is not writable.").arg(arguments.at(1));
        }
        else if (!file.exists() && (!dir.exists() || !dir.isDir() || !dir.isWritable())) {
            return tr("Cannot create file %1.").arg(arguments.at(1));
        }
    }
    return QString();
}

QList< Rdfs::CommandParameter > RdfsDotExport::arguments() const
{
    Rdfs::CommandParameter predicate, file;
    predicate.name = tr("predicate");
    predicate.help = tr("The predicate to export the graph for.");
    predicate.examples << "sp" << "sc";
    predicate.required = true;

    file.name = tr("file");
    file.help = tr("The filename to write to.");
    file.examples << "graph.dot";
    file.required = true;

    return QList<CommandParameter>() << predicate << file;
}

RdfsDotExport::RdfsDotExport(Rdfs::RdfsUpdate* parent): RdfsCommand(parent)
{

}

RdfsDotExport::~RdfsDotExport()
{

}

#include "rdfsdotexport.moc"
