/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "rdfsexport.h"

#include <QFileInfo>
#include <QTextStream>
#include <stdio.h>

#include "../rdfsupdate.h"
#include "../rdfscursor.h"
#include "../rdfslogger.h"
#include <QDir>


using namespace Rdfs;

RdfsExport::RdfsExport(RdfsUpdate* parent): RdfsCommand(parent)
{

}

RdfsExport::~RdfsExport()
{

}

bool RdfsExport::exec(const QStringList& arguments)
{
    // First load table argument (required to open cursor)
    RdfsTable table;
    if (arguments.size() > 2) {
        if (arguments.at(2).compare("original", Qt::CaseInsensitive) == 0) {
            table = RDFS_ORIGINAL;
        }
        else if (arguments.at(2).compare("update", Qt::CaseInsensitive) == 0) {
            table = RDFS_UPDATE;
        }
    }
    else {
        table = RDFS_ORIGINAL;
    }

    // No first open cursor, to make sure we have at least something to export
    Triple t; // Empty triple -> request all triples
    RdfsCursor *c = framework()->store()->cursor(t, table);
    if (!c->open()) {
        RdfsLogger::instance()->error(tr("No triples found in given table - nothing to export."));\
        delete c;
        return false;
    }

    // Open file
    QFile file;
    if (arguments.at(0) != "-") {
        file.setFileName(arguments.at(0));
        file.open(QIODevice::Truncate | QIODevice::WriteOnly | QIODevice::Text);
    }
    else {
        // Open stdout -> bypass RdfsLogger class preferrably
        file.open(stdout, QIODevice::WriteOnly | QIODevice::Text);
    }
    QTextStream stream(&file);

    // Load separator arguments
    QString sep = (arguments.size() > 1) ? arguments.at(1) : " ";

    // Export data
    while (!c->eof()) {
        t = c->current();
        stream << t.subject << sep << t.predicate << sep << t.object << endl;
        c->next();
    }
    c->close();
    file.close();
    delete c;
    return true;
}

QString RdfsExport::checkArguments(const QStringList& arguments) const
{
    if (arguments.at(0).isEmpty()) {
        return tr("No filename was given.");
    }

    if (arguments.at(0) != "-") {
        // Check if file is writable
        QFileInfo file(arguments.at(0));
        QFileInfo dir(file.path());

        if (file.isDir()) {
            return tr("File %1 is a directory.").arg(arguments.at(0));
        }
        else if (file.exists() && !file.isWritable()) {
            return tr("File %1 is not writable.").arg(arguments.at(0));
        }
        else if (!file.exists() && (!dir.exists() || !dir.isDir() || !dir.isWritable())) {
            return tr("Cannot create file %1.").arg(arguments.at(0));
        }
    }

    if (arguments.size() == 3) {
        // Check table argument
        if (arguments.at(2).compare("original", Qt::CaseInsensitive) != 0 &&
            (arguments.at(2).compare("update", Qt::CaseInsensitive) != 0)) {
            return tr("Unknown table %1 given.").arg(arguments.at(3));
        }
    }
    return QString();
}

QList< Rdfs::CommandParameter > RdfsExport::arguments() const
{
    QList<CommandParameter> list;

    CommandParameter file;
    file.name = tr("file");
    file.help = tr("The filename to export the data to. Use '-' for stdout.");
    file.examples << tr("/home/user/export/file.rdfsupdate") << "-";
    file.required = true;
    list << file;

    CommandParameter sep;
    sep.name = tr("separator");
    sep.help = tr("The separator used to separate the atoms in a triple. Defaults to a single space.");
    sep.examples << "\" \"" << ",";
    sep.required = false;
    list << sep;

    CommandParameter table;
    table.name = tr("table");
    table.examples << "original";
    table.examples << "update";
    table.help = tr("Table to use for this command");
    table.required = false;
    list << table;

    return list;
}

#include "rdfsexport.moc"
