/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

// See: http://www.boost.org/doc/libs/1_41_0/libs/graph/doc/plod_generator.html

#include "rdfsgraphgen.h"

#include <QSet>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/plod_generator.hpp>
#include <boost/random/linear_congruential.hpp>

#include "../triple.h"
#include "../rdfsupdate.h"
#include "../rdfslogger.h"


struct IndexVertex {
    quint32 index;
};

typedef boost::adjacency_list<boost::slistS, boost::slistS, boost::directedS, IndexVertex> Graph;
typedef boost::graph_traits<Graph>::out_edge_iterator OutEdgeIterator;
typedef boost::graph_traits<Graph>::edge_descriptor Edge;
typedef boost::graph_traits<Graph>::vertex_iterator VertexIterator;
typedef boost::graph_traits<Graph>::vertex_descriptor Vertex;
typedef boost::plod_iterator<boost::minstd_rand, Graph> SFGen;

using namespace Rdfs;

bool RdfsGraphGen::exec(const QStringList& arguments)
{
    bool ok;
    const QString predicate = arguments.at(0);
    const QString prefix = arguments.at(1);
    const QString type = arguments.at(2);
    const int n = arguments.at(3).toInt(&ok);
    const double a = (arguments.size() >= 5) ? arguments.at(4).toDouble(&ok) : 2.72;
    const double b = (arguments.size() >= 6) ? arguments.at(5).toDouble(&ok) : 100;

    // Create graph
    RdfsLogger::instance()->log(tr("Generating graph..."));
    // Seed with random number. If our random is seeded with constant prior to this, then
    // we'll still always have the same seed. Otherwise, the results will differ :)
    boost::minstd_rand gen;
    gen.seed(qrand());
    Graph g(SFGen(gen, n, a, b, false), SFGen(), n);


    // Assign indexes
    RdfsLogger::instance()->log(tr("Assigning vertex indices..."));
    VertexIterator vi, vend;
    quint32 i = 0;
    for (boost::tie<VertexIterator, VertexIterator>(vi, vend) = boost::vertices(g); vi != vend; ++vi) {
        g[*vi].index = i;
        ++i;
    }

    // Load graph into database
    RdfsLogger::instance()->log(tr("Loading graph into database..."));
    int count = 0;
    framework()->store()->startBulkLoad();
    for (boost::tie<VertexIterator, VertexIterator>(vi, vend) = boost::vertices(g); vi != vend; ++vi) {
        QSet<quint32> edges;
        Vertex v1 = *vi;

        // Get (unique) outgoing edges
        OutEdgeIterator ei, eend;
        for (boost::tie<OutEdgeIterator, OutEdgeIterator>(ei, eend) = boost::out_edges(v1, g); ei != eend; ++ei) {
            Edge e = *ei;
            Vertex v2 = e.m_target;

            if (edges.contains(g[v2].index)) {
                RdfsLogger::instance()->debug(tr("Duplicate edge %1 -> %2").arg(g[v1].index).arg(g[v2].index));
                continue;
            }

            Rdfs::Triple t;
            t.subject = prefix + QString::number(g[v1].index);
            t.predicate = predicate;
            t.object = prefix + QString::number(g[v2].index);
            framework()->store()->insertTriple(t);

            edges.insert(g[v2].index);

            count++;
        }
    }
    RdfsLogger::instance()->log(tr("Inserted %1 edges...").arg(count));

    RdfsLogger::instance()->log(tr("Loading type statements..."));
    for (int i = 0; i < n; ++i) {
        Triple t;
        t.subject = prefix + QString::number(i);
        t.predicate = "type";
        t.object = type;
        framework()->store()->insertTriple(t);
    }
    RdfsLogger::instance()->debug("Inserted type statements");

    RdfsLogger::instance()->log(tr("Wrapping up..."));
    framework()->store()->endBulkLoad();

    RdfsLogger::instance()->log(tr("Done. Please use \"killcycles %1\" to remove any cycles in the graph.").arg(predicate));
    return true;
}

QString RdfsGraphGen::checkArguments(const QStringList& arguments) const
{
    bool ok;
    int n = arguments.at(3).toInt(&ok);
    if (!ok) {
        return tr("The value for n must be an integer");
    }

    float a, b;
    if (arguments.size() >= 5) {
        a = arguments.at(4).toDouble(&ok);
        if (!ok) {
            return tr("The value for alpha must be a number");
        }
    }

    if (arguments.size() >= 6) {
        b = arguments.at(5).toDouble(&ok);
        if (!ok) {
            return tr("The value for beta must be a number");
        }
    }
    return QString();
}

QList< Rdfs::CommandParameter > RdfsGraphGen::arguments() const
{
    CommandParameter predicate, prefix, type, n, alpha, beta;
    predicate.name = tr("predicate");
    predicate.help = tr("The predicate to use as edge");
    predicate.examples << "sp" << "sc";
    predicate.required = true;

    prefix.name = tr("prefix");
    prefix.help = tr("The naming prefix for vertices");
    prefix.examples << "Class" << "Prop";
    prefix.required = true;

    type.name = tr("type");
    type.help = tr("The type to assign to the vertices");
    type.examples << "Class" << "Property";
    type.required = true;

    n.name = tr("n");
    n.help = tr("Amount of vertices to place in graph");
    n.examples << "100";
    n.required = true;

    alpha.name = tr("alpha");
    alpha.help = tr("The value controlling how steeply the curve drops off. A larger value indicates a steeper curve.");
    alpha.examples << "2.72";
    alpha.required = false;

    beta.name = tr("beta");
    beta.help = tr("The value controlling the average degree of vertices.");
    beta.examples << "100";
    beta.required = false;

    return QList<CommandParameter>() << predicate << prefix << type << n << alpha << beta;

}


RdfsGraphGen::RdfsGraphGen(RdfsUpdate *parent)
  : RdfsCommand(parent)
{

}

RdfsGraphGen::~RdfsGraphGen()
{

}

#include "rdfsgraphgen.moc"
