/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "rdfsloadfile.h"

#include <QFileInfo>
#include <QTextStream>

#include "../rdfslogger.h"
#include "../triple.h"
#include "../rdfsupdate.h"

using namespace Rdfs;

RdfsLoadFile::RdfsLoadFile(RdfsUpdate* parent): RdfsCommand(parent)
{
}

bool RdfsLoadFile::exec(const QStringList& arguments)
{
    const QString sep = (arguments.size() > 1) ? arguments.at(1) : " ";
    RdfsTable table =  RDFS_ORIGINAL;
    #ifndef NDEBUG
    int i = 0;
    #endif

    // Load table argument
    if (arguments.size() > 2) {
        if (arguments.at(2).compare("original", Qt::CaseInsensitive) == 0) {
            table = RDFS_ORIGINAL;
        }
        else if (arguments.at(2).compare("update", Qt::CaseInsensitive) == 0) {
            table = RDFS_UPDATE;
        }
    }

    QFile f(arguments.at(0));
    if (!f.open(QIODevice::ReadOnly | QIODevice::Text)) {
        RdfsLogger::instance()->error(tr("Error opening file \"%1\"").arg(f.fileName()));
        return false;
    }

    QTextStream data(&f);
    framework()->store()->startBulkLoad();
    while (!data.atEnd()) {
        QString line = data.readLine();
        QStringList fields = line.split(sep, QString::KeepEmptyParts);
        if (fields.length() != 3) {
            RdfsLogger::instance()->error(tr("Incorrectly formatted line, could not load triple:\n %1").arg(line));
            continue;
        }

        Triple t;
        t.subject = fields.at(0);
        t.predicate = fields.at(1);
        t.object = fields.at(2);

        framework()->store()->insertTriple(t, table);

        #ifndef NDEBUG
        ++i;
        if ((i % 1000) == 0) {
            RdfsLogger::instance()->debug(tr("Inserted %1 triples").arg(i));
        }
        #endif
    }
    f.close();
    framework()->store()->endBulkLoad();
    return true;
}

QString RdfsLoadFile::checkArguments(const QStringList& arguments) const
{
    QString error;

    // Check file
    QFileInfo f(arguments.at(0));
    if (!f.exists()) {
        error = tr("The file \"%1\" could not be found.").arg(f.absoluteFilePath());
    }
    else if (!f.isFile()) {
        error = tr("File \"%1\" is not a file.").arg(f.absoluteFilePath());
    }

    // No need to check separator

    // Check table, if given
    if ((arguments.count() == 3) && (arguments.at(2).compare("original", Qt::CaseInsensitive) != 0 &&
        (arguments.at(2).compare("update", Qt::CaseInsensitive) != 0))) {
        error = tr("Unknown table %1 given.").arg(arguments.at(2));
    }

    return error;
}

bool RdfsLoadFile::requireConnection() const
{
    return true;
}

QList<CommandParameter> RdfsLoadFile::arguments() const
{
    CommandParameter file;
    file.name = tr("filename");
    file.examples << tr("triples.db");
    file.help = tr("The filename containing triples (one per line, form \"subject predicate object\")");
    file.required = true;

    CommandParameter sep;
    sep.name = tr("separator");
    sep.examples << ",";
    sep.help = tr("The separator between subject, predicate and object in the triples. Defaults to a space");
    sep.required = false;

    CommandParameter table;
    table.name = tr("table");
    table.examples << "original";
    table.examples << "update";
    table.help = tr("Table to use for this command");
    table.required = false;

    return (QList<CommandParameter>() << file << sep << table);
}

QStringList RdfsLoadFile::aliases() const
{
    return QStringList() << "l";
}

QString RdfsLoadFile::description() const
{
    return tr("Load a formatted file with triples into the database.");
}

QString RdfsLoadFile::name() const
{
    return "load";
}

Rdfs::RdfsLoadFile::~RdfsLoadFile()
{
}

#include "rdfsloadfile.moc"
