/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "rdfsontologylinks.h"

#include <QtGlobal>
#include <stdlib.h>

#include "../triple.h"
#include "../rdfsupdate.h"
#include "../rdfsbackend.h"
#include "../rdfscursor.h"
#include "../rdfslogger.h"

using namespace Rdfs;

bool RdfsOntologyLinks::exec(const QStringList& arguments)
{
    bool ok;
    QString propType = (arguments.size() > 0) ? arguments.at(0) : "Property";
    QString classType = (arguments.size() > 1) ? arguments.at(1) : "Class";
    double zeroDegreePct = (arguments.size() > 2) ? arguments.at(2).toDouble(&ok) : 0.5;
    double selfLoopPct = (arguments.size() > 3) ? arguments.at(3).toDouble(&ok) : 0.126;
    QList<QString> classes;

    // Collect classes
    RdfsLogger::instance()->log(tr("Collecting classes..."));
    Triple csearch;
    csearch.subject = QString();
    csearch.predicate = "type";
    csearch.object = classType;
    RdfsCursor *c = framework()->store()->cursor(csearch);

    if (!c->open()) {
        delete c;
        RdfsLogger::instance()->error(tr("No atoms with type %1 (classes) found").arg(classType));
        return false;
    }

    while (!c->eof()) {
        Triple t = c->current();
        if (chance(zeroDegreePct)) {
            // Skip class - need zero degree
            continue;
        }
        classes.append(t.subject);
        if (classes.size() % 1000 == 0) {
            RdfsLogger::instance()->log(tr("Collected %1 classes so far...").arg(classes.size()));
        }
        c->next();
    }
    delete c;

    // Assigning domains/properties
    RdfsLogger::instance()->log(tr("Assigning domains and ranges..."));
    Triple psearch;
    psearch.subject = QString();
    psearch.predicate = "type";
    psearch.object = propType;
    c = framework()->store()->cursor(psearch);

    if (!c->open()) {
        delete c;
        RdfsLogger::instance()->error(tr("No atoms with type %1 (properties) found").arg(propType));
        return false;
    }

    framework()->store()->startBulkLoad();
    while (!c->eof()) {
        Triple t = c->current();
        QString domain, range;

        domain = classes.at(qrand() % classes.size());
        if (chance(selfLoopPct)) {
            // Skip class - need zero degree
            range = domain;
        }
        else {
            range = classes.at(qrand() % classes.size());
        }

        // Insert data
        Triple tdomain, trange;
        tdomain.subject = t.subject;
        tdomain.predicate = "dom";
        tdomain.object = domain;
        framework()->store()->insertTriple(tdomain);

        trange.subject = t.subject;
        trange.predicate = "range";
        trange.object = range;
        framework()->store()->insertTriple(trange);

        RdfsLogger::instance()->debug(tr("Assigned: %1 -%2-> %3").arg(domain, t.subject, range));

        c->next();
    }
    framework()->store()->endBulkLoad();
    delete c;

    return true;
}

bool RdfsOntologyLinks::chance(float p)
{
    int rgn = qrand();
    return (rgn < qRound(p * (float)RAND_MAX));
}

QList< Rdfs::CommandParameter > RdfsOntologyLinks::arguments() const
{
    CommandParameter propType, classType, zeroDegreePct, selfLoopPct;
    propType.name = tr("property");
    propType.help = tr("The type of properties. All properties should have an existing \"Name type Property\" statement in the database, "
                       "with Property being the type for properties.");
    propType.examples << "Property";
    propType.required = false;

    classType.name = tr("classType");
    classType.help = tr("The type of classes. All classes should have an existing \"Name type Class\" statement in the database, "
                       "with Class being the type for classes.");
    classType.examples << "Class";
    classType.required = false;

    zeroDegreePct.name = tr("zerodegreepct");
    zeroDegreePct.help = tr("The fraction of classes (0-1) which should not be assigned as domain or range of properties at all.");
    zeroDegreePct.examples << "0.5";
    zeroDegreePct.required = false;

    selfLoopPct.name = tr("selflooppct");
    selfLoopPct.help = tr("The fraction of properties that should construct a self-loop, i.e. an equal domain and range.");
    selfLoopPct.examples << "0.126";
    selfLoopPct.required = false;

    return QList<CommandParameter>() << propType << classType << zeroDegreePct << selfLoopPct;
}

QString RdfsOntologyLinks::checkArguments(const QStringList& arguments) const
{
    if (arguments.size() > 2) {
        // Check zeroDegreePct
        bool ok;
        double zeroDegreePct = arguments.at(2).toDouble(&ok);

        if (!ok) {
            return tr("The fraction of classes with a degree of zero is not numeric.");
        }
        if (0 > zeroDegreePct || 1 < zeroDegreePct) {
            return tr("The fraction of classes with a degree of zero must be between 0 and 1.");
        }
    }
    if (arguments.size() > 3) {
        // Check selfLoopPct
        bool ok;
        double selfLoopPct = arguments.at(3).toDouble(&ok);

        if (!ok) {
            return tr("The fraction of self-loop properties is not numeric.");
        }
        if (0 > selfLoopPct || 1 < selfLoopPct) {
            return tr("The fraction of self-loop properties must be between 0 and 1.");
        }
    }
    return QString();
}

RdfsOntologyLinks::RdfsOntologyLinks(Rdfs::RdfsUpdate* parent): RdfsCommand(parent)
{

}

RdfsOntologyLinks::~RdfsOntologyLinks()
{

}

#include "rdfsontologylinks.moc"
