/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef RDFS_RDFSUNCHASE_H
#define RDFS_RDFSUNCHASE_H

#include <QSet>
#include <QList>

#include "rdfstriplecommand.h"

typedef QMap<QString, QString> VariableInstances;

namespace Rdfs {
class RdfsUnchase : public Rdfs::RdfsTripleCommand {
    Q_OBJECT
private:
    /**
     * Returns a set of triples which derive the given triple, according to the
     * (deterministic) custom rules. The triples themselves are the key of the map.
     * If a rule which matches, instantiates any of t's (bound) variables, this
     * variable/instantiation mapping is given in the value of the map.
     */
    QMap<Triple, VariableInstances> sources(Triple t, QList<Rule> rules, RdfsTable table);
    /**
     * Find instances of a triple in the given table. This returns the set of
     * triples which have a valid instantiation for the (bound) variables in the
     * triple. The variables which were bound are in the value of the mapping.
     */
    QMap<Triple, VariableInstances> tripleInstances(Triple t, RdfsTable table);

    /**
     * Check if a term of a target triple of a rule is a valid instantiation of
     * another triple. If necessary, bind local and/or global variables.
     * @param t1 The term of the triple which needs to be matched
     * @param t2 The term of the possible instantiation
     * @param localVars A mapping of variables bound in the current rule
     * @param boundVars A mapping of the initial triple variable values
     */
    bool checkBinding(QString t1, QString t2, VariableInstances *localVars, VariableInstances *boundVars);

    /**
     * Instantiate a triple with the given variables. This replaces the bound
     * variables ($1, $2 etc) in the triple with the corresponding values from @p vars.
     */
    Triple instantiate(Triple t, VariableInstances vars);

    /**
     * Return @p t with all variables ($1, $2 etc) as blanks.
     */
    Triple clearVars(Triple t);

    /**
     * Merge two variable instance mappings.
     */
    VariableInstances mergeVarInstances(VariableInstances i1, VariableInstances i2);

    /**
     * Show a variable map as string. Used during debugging.
     */
    QString variableMapToString(VariableInstances instances);

    /**
     *  The current amount of bound variables. Use and increment this value to
     *  create a new variable
     */
    int m_boundVarCount;

public:

    RdfsUnchase(RdfsUpdate* parent);
    virtual ~RdfsUnchase();

    virtual bool exec(const QStringList& arguments);
    virtual QString checkArguments(const QStringList& arguments) const;
    virtual QStringList aliases() const;
    virtual QString description() const;
    virtual QString name() const;

    QMap<Triple, VariableInstances> unchase(Triple t, QList<Rule> rules, RdfsTable table, QSet<Triple> processed = QSet<Triple>());
};

}

#endif // RDFS_RDFSUNCHASE_H
