/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef RDFS_RDFSBACKEND_H
#define RDFS_RDFSBACKEND_H

#include <QStringList>
#include <QList>
#include <QMap>

#include "triple.h"
#include "cachestats.h"

namespace Rdfs {

enum RdfsTable {
    RDFS_ORIGINAL = 0,
    RDFS_UPDATE   = 1,
    RDFS_RULE     = 2 // Internal use only: not available as argument for commands
};

enum RdfsCapabilities {
    RDFS_NONE     = 0x00,
    RDFS_INFER_SP = 0x01,
    RDFS_INFER_SC = 0x02,
    RDFS_CURSOR   = 0x04,
    RDFS_CACHESTATS = 0x08,
    RDFS_TRANSITIVE = 0x10,
};

enum RdfsRole {
    RDFS_SUBJECT = 0,
    RDFS_PREDICATE = 1,
    RDFS_OBJECT = 2,
};

class RdfsCursor;

/**
 * This is an interface which can be implemented by a backend to store the RDFS
 * tables in. All abstract virtuals need to be reimplemented by a backend to
 * be functional.
 */
class RdfsBackend
{
protected:
    /**
     * Get the actual name of a specific table, given by an RdfsTable.
     * @param t The RdfsTable enum value to get the table name of.
     */
    QString tableName(RdfsTable t);

    /**
     * Constructor. Reimplement the actual constructor; the RdfsBackend itself
     * cannot be instantiated.
     */
    RdfsBackend();

public:
    virtual ~RdfsBackend();

    /**
     * Check if the connection is open.
     */
    virtual bool isOpen() = 0;

    /**
     * Close the open connection.
     */
    virtual void close() = 0;

    /**
     * Get the capabilites of this backend.
     */
    virtual int capabilities() const = 0;

    /**
     * Get the keys of options this backend supports.
     */
    virtual QStringList options() const { return QStringList(); };

    /**
     * Set an option for this back-end. The keys are available using
     * options().
     *
     * @param key The key of the option to set
     * @param value The value of the option.
     * @return Whether the option was set successfully or not.
     * @see options()
     */
    virtual bool setOption(const QString &key, const QString &value) { return false; };


    /**
     * Get the maximal transitive length supported by the back-end. The predicates which
     * are considered transitive are given by transitivePredicates(). This distance
     * indicates the amount of extra triples which are returned by cursors and when using
     * e.g. findTriples().
     *
     * Example, say predicate p is transitive and the distance is 2. Consider a graph
     * with (a p b), (b p c), (c p d), (e p f), (f p g) and (g p h). Adding in (d p e)
     * will implicitly also add (c p e), (b p e), (d p f) and (d p g) but not (a p e) or
     * (d p h).
     *
     * If this returns zero, no inference for transitive predicates is done by the back-end.
     */
    virtual unsigned char transitiveDistance() const { return 0; };


    /**
     * Get the transitive predicates. These are the atoms which, when used as predicate,
     * will trigger the implicit creation of additional triples. See transitiveDistance()
     * for the amount of implicitly created triples.
     */
    virtual QStringList transitivePredicates() const { return QStringList(); };

    /**
     * Start bulk insertion of triples. Default implementation does nothing and assumes
     * no special optimizations are made for large inserts. If this is supported by the
     * backend, insertTriple() will use the bulk insert until endBulkLoad() is called.
     *
     * @see insertTriple
     * @see endBulkLoad
     */
    virtual void startBulkLoad() {};

    /**
     * End bulk insertion of triple.
     *
     * @see insertTriple
     * @see startBulkLoad
     */
    virtual void endBulkLoad() {};

    /**
     * Insert a new triple in the database
     * @param t The triple to insert. Must be a complete triple (no blanks).
     * @param table The table to check in (Original or Update)
     */
    virtual void insertTriple(const Triple &t, RdfsTable table = RDFS_ORIGINAL) = 0;

    /**
     * Check how many rows match the given triple. If a triple field (subject,
     * predicate or object) is empty, it is not checked on. An empty triple
     * (all empty strings) gives the amount of total rows in the given table.
     * @param t The triple to check (may contain blanks).
     * @param table The table to check in (Original or Update)
     */
    virtual int checkTriple(const Triple &t, RdfsTable table = RDFS_ORIGINAL) = 0;

    /**
     * Check if a triple (with possible blank atoms) exists in the graph or not.
     * This may be more efficient than using checkTriple(), since checkTriple() may
     * actually fetch or at least loop over all results rather than just performing the
     * check.
     *
     * The default implementation simply attempts to open a cursor with the given
     * query. If this succeeds, the triple exists, otherwise it doesn't.
     */
    virtual bool hasTriple(const Triple &t, RdfsTable table = RDFS_ORIGINAL);

    /**
     * Find the triples which match the given triple @p t (with optionally blank
     * subject, predicate and/or object, which will then be ignored in the query).
     * The triples are returned in a list.
     * @param t The triple to check (may contain blanks).
     * @param table The table to check in (Original or Update)
     */
    virtual QList<Triple> findTriples(const Rdfs::Triple& t, RdfsTable table) = 0;

    /**
     * Remove a triple from the database
     * @param t The triple to remove. Must be a complete triple (no blanks).
     * @param table The table to check in (Original or Update)
     */
    virtual void removeTriple(const Triple &t, RdfsTable table = RDFS_ORIGINAL) = 0;

    /**
     * Insert a new rule. The triples are located or added in the RDFS_RULE table.
     * Next, the actual rule is added to the system created.
     * @param r The rule to add (pair of triples, the first being the source and
     *          the second being the target triple).
     */
    virtual void insertRule(Rule r) = 0;

    /**
     * Insert a new rule. The triples are located or added in the RDFS_RULE table.
     * Next, the insertRule(int, int) method is called with the row ID's of the
     * triples. Do not override, but override insertRule(int, int) instead.
     * @param t1 The source triple of the rule.
     * @param t2 The destination triple of the rule.
     */
    void insertRule(Triple t1, Triple t2);

    /**
     * Retrieve all deterministic tuple-generating dependencies in a list.
     */
    virtual QList<Rule> rules() = 0;

    /**
     * Get a cursor for a specific query. If the back-end supports this,
     * a cursor will be returned. This can be used to iterate over the results,
     * without having to load the entire list in memory first.
     *
     * If the back-end does not support cursors, the query result using findTriples()
     * will be kept locally, and a new RdfsListCursor will be returned by the default
     * implementation.
     *
     * Note: if you reimplement this class, do NOT call the base implementation, as that
     * will use findTriples() to retrieve all the results and store that in memory.
     * You probably do not need this.
     *
     * @param q The query triple. Triples returned by the cursor will match this
     *          triple.
     * @param table The table to look up triples in.
     * @see capabilities()
     */
    virtual RdfsCursor *cursor(Triple q, RdfsTable table = RDFS_ORIGINAL);

    /**
     * Remove a rule by its triples. If the given triples do not form a defined
     * rule, nothing happens.
     * @param r The rule (triple pair) to remove.
     */
    virtual void removeRule(Rule r) = 0;

    /**
     * Remove a rule by its triples. If the given triples do not form a defined
     * rule, nothing happens.
     * @param t1 The source triple of the rule.
     * @param t2 The target triple of the rule.
     */
    void removeRule(Triple t1, Triple t2);

    /**
     * Truncate a table. This removes all triples in the given table.
     * @param table The table to clear (Original or Update).
     */
    virtual void truncate(RdfsTable table = RDFS_ORIGINAL) = 0;

    /**
     * Get page statistics of the database. These are actual stats for
     * the entire database. If the backend does not support this, the
     * default implementation will return a PageStats value with all
     * 0 values. Otherwise, more meaningful elements should be returned,
     * depending ont he backend. See capabilities() to inspect if the
     * page statistics are supported by the backend.
     *
     * @return Statistics of the page size, hit/miss ratio etc. of the
     *         database.
     */
    virtual CacheStats statistics() const;
};
}

#endif // RDFS_RDFSBACKEND_H
