/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "rdfsbdb.h"
#include "rdfslogger.h"
#include <db.h>
#include <stdlib.h>
#include <QDir>
#include <QFile>
#include <QFileInfo>
#include "rdfsbdbcursor.h"

using namespace Rdfs;

RdfsBdb::RdfsBdb(const QString &dir, const QString &file, const QString &incomingFile, const QString &outgoingFile, const QString &edgeFile,
                 const QString &cacheSize, const QString &pageSize)
    : RdfsBackend()
{
    m_db_env = NULL;
    m_db_primary = NULL;
    m_db_in = NULL;
    m_db_out = NULL;
    m_db_edge = NULL;

#ifdef DB_MULTIPLE_RECNO_WRITE_INIT
    m_bulkLoad = false;
    m_bulkPointer = NULL;
    m_bulkData = NULL;
    m_nextRecno = 0;
#endif

    if (file.isEmpty() || incomingFile.isEmpty() || outgoingFile.isEmpty() || edgeFile.isEmpty()) {
        return;
    }

    // Remove old cache files. This (a) clears cache, (b) allows opening db with flexible cache sizes
    QDir envdir(dir);
    QFileInfoList cacheFiles = envdir.entryInfoList(QStringList() << "__db.*", QDir::Files | QDir::NoDotAndDotDot);
    foreach(QFileInfo f, cacheFiles) {
        RdfsLogger::instance()->debug(QString("Remove stale cache file: %1").arg(f.absoluteFilePath()));
        QFile::remove(f.absoluteFilePath());
    }

    u_int32_t flags, env_flags; // Database/environment open flags
    int ret; // Return value

    // Create & open environment
    if (db_env_create(&m_db_env, 0) != 0) {
        return;
    }

    if (!cacheSize.isEmpty()) {
        setCacheSize(cacheSize);
    }

    env_flags = DB_CREATE | DB_INIT_MPOOL;
    if ((ret = m_db_env->open(m_db_env, dir.toUtf8().constData(), env_flags, 0)) != 0) {
        RdfsLogger::instance()->error(QObject::tr("Error opening database environment at %1 (error %2)").arg(dir, QString::number(ret)));
        return;
    }

    if (db_create(&m_db_primary, m_db_env, 0) != 0) {
        close();
        return;
    }
    if (db_create(&m_db_in, m_db_env, 0) != 0) {
        close();
        return;
    }
    if (db_create(&m_db_out, m_db_env, 0) != 0) {
        close();
        return;
    }
    if (db_create(&m_db_edge, m_db_env, 0) != 0) {
        close();
        return;
    }

    // Allow dupes in secondary databases
    m_db_in->set_flags(m_db_in, DB_DUP);
    m_db_out->set_flags(m_db_out, DB_DUP);
    m_db_edge->set_flags(m_db_edge, DB_DUP);

    if (!pageSize.isEmpty()) {
        setPageSize(pageSize);
    }

    // Create non-existing dbs
    flags = DB_CREATE;

    // Open databases
    if (m_db_primary->open(m_db_primary, NULL, file.toLatin1().constData(), "primary", DB_RECNO, flags, 0) != 0) {
        close();
        return;
    }
    if (m_db_in->open(m_db_in, NULL, incomingFile.toLatin1().constData(), "in", DB_HASH, flags, 0) != 0) {
        close();
        return;
    }
    if (m_db_out->open(m_db_out, NULL, outgoingFile.toLatin1().constData(), "out", DB_HASH, flags, 0) != 0) {
        close();
        return;
    }
    if (m_db_edge->open(m_db_edge, NULL, edgeFile.toLatin1().constData(), "edge", DB_HASH, flags, 0) != 0) {
        close();
        return;
    }

    // And associate secondary dbs
    m_db_primary->associate(m_db_primary, NULL, m_db_in, RdfsBdb::incomingKey, 0);
    m_db_primary->associate(m_db_primary, NULL, m_db_out, RdfsBdb::outgoingKey, 0);
    m_db_primary->associate(m_db_primary, NULL, m_db_edge, RdfsBdb::edgeKey, 0);
}

int RdfsBdb::incomingKey(DB* db_handle, const DBT* pkey, const DBT* data, DBT* skey)
{
    // Key is object
    loadKey(data, RDFS_OBJECT, skey);
    return 0;
}

int RdfsBdb::outgoingKey(DB* db_handle, const DBT* pkey, const DBT* data, DBT* skey)
{
    // Key is subject
    loadKey(data, RDFS_SUBJECT, skey);
    return 0;
}

int RdfsBdb::edgeKey(DB* db_handle, const DBT* pkey, const DBT* data, DBT* skey)
{
    // Key is predicate
    loadKey(data, RDFS_PREDICATE, skey);
    return 0;
}

void RdfsBdb::loadKey(const DBT* data, TripleElement index, DBT *skey)
{
    memset(skey, 0, sizeof(DBT));

    // Decode 'data' - lengths first
    qint16 lengths[3];
    u_int8_t *p = (u_int8_t *)data->data;
    memcpy(&(lengths[0]), p, sizeof(qint16)); // TODO This may be possible with one memcpy() ? - Check performance
    p += sizeof(qint16);
    memcpy(&(lengths[1]), p, sizeof(qint16));
    p += sizeof(qint16);
    memcpy(&(lengths[2]), p, sizeof(qint16));
    p += sizeof(qint16);

    int i = 0;
    if (index == RDFS_PREDICATE) {
        p += lengths[0];
        i = 1;
    }
    else if (index == RDFS_OBJECT) {
        p += lengths[0] + lengths[1];
        i = 2;
    }

    skey->data = p;
    skey->size = lengths[index];
}

DBT RdfsBdb::tripleToDbt(Triple t)
{
    QByteArray uSubject = t.subject.toUtf8();
    QByteArray uPredicate = t.predicate.toUtf8();
    QByteArray uObject = t.object.toUtf8();

    // TODO Make array int[3], copy all in one go? - Check performance
    qint16 sLength = uSubject.length();
    qint16 pLength = uPredicate.length();
    qint16 oLength = uObject.length();

    int length = uSubject.length() + uPredicate.length() + uObject.length();
    int dataLength = length*sizeof(char) + 3*sizeof(qint16);

    char *data = (char *)malloc(dataLength); // 3 length indicators (int16)

    memset(data, 0, dataLength);

    // Write lengths
    memcpy(data, &sLength, sizeof(qint16));
    data += sizeof(qint16);
    memcpy(data, &pLength, sizeof(qint16));
    data += sizeof(qint16);
    memcpy(data, &oLength, sizeof(qint16));
    data += sizeof(qint16);

    // Write data
    memcpy(data, uSubject, sLength * sizeof(char));
    data += sLength * sizeof(char);
    memcpy(data, uPredicate, pLength * sizeof(char));
    data += pLength * sizeof(char);
    memcpy(data, uObject, oLength * sizeof(char));
    data += oLength * sizeof(char);

    // Reset pointer to beginning
    data -= dataLength;

    // Construct result
    DBT result;
    memset(&result, 0, sizeof(DBT));
    result.data = data;
    result.size = dataLength;

    return result;
}

Triple RdfsBdb::dbtToTriple(DBT *data)
{
    Triple t;

    // Decode 'data' - lengths first
    qint16 *lengths = new qint16[3];
    char *p = (char *)data->data;
    memcpy(&(lengths[0]), p, sizeof(qint16)); // TODO This may be possible with one memcpy() ? - Check performance
    p += sizeof(qint16);
    memcpy(&(lengths[1]), p, sizeof(qint16));
    p += sizeof(qint16);
    memcpy(&(lengths[2]), p, sizeof(qint16));
    p += sizeof(qint16);

    t.subject = QString::fromUtf8(p, lengths[0]);
    p += lengths[0];
    t.predicate = QString::fromUtf8(p, lengths[1]);
    p += lengths[1];
    t.object = QString::fromUtf8(p, lengths[2]);

    delete[] lengths;

    return t;
}


RdfsBdb::~RdfsBdb()
{
    if (isOpen()) {
        close();
    }
}


void RdfsBdb::truncate(Rdfs::RdfsTable table)
{
    m_db_primary->truncate(m_db_primary, NULL, NULL, 0);
}

void RdfsBdb::removeRule(Rdfs::Rule r)
{

}

QList< Rdfs::Rule > RdfsBdb::rules()
{
    return QList<Rdfs::Rule>();
}

void RdfsBdb::insertRule(Rdfs::Rule r)
{

}

void RdfsBdb::removeTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table)
{
    DBC *in_cursor, *out_cursor, *edge_cursor, *join_cursor;

    int flags;
    int ret = findTripleCursor(t, table, &join_cursor, &in_cursor, &out_cursor, &edge_cursor, &flags);
    if (ret != 0) {
        RdfsLogger::instance()->error(QObject::tr("Error loading cursor for deleting triples (error code %1)").arg(QString::number(ret)));
        return;
    }

    // Init key/data pairs
    DBT key, data;
    memset(&key, 0, sizeof(DBT));
    memset(&data, 0, sizeof(DBT));

    // Gather keys to remove (join cursor doesn't support c_del() )
    QList<DBT> keys;
    while ((ret = join_cursor->c_get(join_cursor, &key, &data, flags)) == 0) {
        DBT keyCopy;
        memset(&keyCopy, 0, sizeof(keyCopy));
        keyCopy.data = malloc(key.size);
        memcpy(keyCopy.data, key.data, key.size);
        keyCopy.size = key.size;
        keys << keyCopy;
    }

    if (ret == DB_NOTFOUND) {
        Q_FOREACH(key, keys) {
            qint32 *d = (qint32 *)key.data;
            ret = m_db_primary->del(m_db_primary, NULL, &key, 0);
            free(key.data);
        }
    }
    else {
        RdfsLogger::instance()->error(QObject::tr("Error deleting triples (error code %1)").arg(QString::number(ret)));
    }

    closeCursors(join_cursor, in_cursor, out_cursor, edge_cursor);
}

QList< Rdfs::Triple > RdfsBdb::findTriples(const Rdfs::Triple& t, Rdfs::RdfsTable table)
{
    DBC *in_cursor, *out_cursor, *edge_cursor, *join_cursor;

    int flags;
    int ret = findTripleCursor(t, table, &join_cursor, &in_cursor, &out_cursor, &edge_cursor, &flags);
    if (ret != 0) {
        return QList<Rdfs::Triple>();
    }

    // Init key/data pairs
    DBT key, data;
    memset(&key, 0, sizeof(DBT));
    memset(&data, 0, sizeof(DBT));

    // Start main get
    QList<Rdfs::Triple> result;
    while ((ret = join_cursor->c_get(join_cursor, &key, &data, flags)) == 0) {
        result << dbtToTriple(&data);
    }

    closeCursors(join_cursor, in_cursor, out_cursor, edge_cursor);

    // Return result
    if (ret == DB_NOTFOUND) {
        return result;
    }
    // Other error
    return QList<Rdfs::Triple>();
}

int RdfsBdb::findTripleCursor(const Triple t, Rdfs::RdfsTable table, DBC** main_cursor, DBC** in, DBC** out, DBC** edge, int *flags)
{
    DBC *carray[4];
    DBT key, data;
    int ret;

    *in = *out = *edge = *main_cursor = NULL;
    int i = 0;
    *flags = 0;
    // Set all indices to NULL for now
    memset(carray, 0, sizeof(carray));
    memset(&key, 0, sizeof(DBT));
    memset(&data, 0, sizeof(DBT));

    if (t.subject.isEmpty() && t.predicate.isEmpty() && t.object.isEmpty()) {
        // Return all triples
        ret = m_db_primary->cursor(m_db_primary, NULL, main_cursor, 0);
        *flags = DB_NEXT;
    }
    else {
        // Start cursors for all set elements, and join them
        if (!t.subject.isEmpty()) {
            ret = secondaryDbCursor(t.subject, m_db_out, out, &key, &data);
            if (ret == 0) {
                carray[i++] = *out;
            }
            else {
                return ret;
            }
        }
        if (!t.predicate.isEmpty()) {
            ret = secondaryDbCursor(t.predicate, m_db_edge, edge, &key, &data);
            if (ret == 0) {
                carray[i++] = *edge;
            }
            else {
                return ret;
            }
        }
        if (!t.object.isEmpty()) {
            ret = secondaryDbCursor(t.object, m_db_in, in, &key, &data);
            if (ret == 0) {
                carray[i++] = *in;
            }
            else {
                return ret;
            }
        }

        // Join cursor
        ret = m_db_primary->join(m_db_primary, carray, main_cursor, 0);
        if (ret != 0) {
            RdfsLogger::instance()->debug("Failed to create join!");
            return ret;
        }
    }
    return ret;
}


void RdfsBdb::closeCursors(DBC* main_cursor, DBC* in, DBC* out, DBC* edge)
{
    // Close cursors
    if (main_cursor != NULL) {
        main_cursor->c_close(main_cursor);
    }
    if (in != NULL) {
        in->c_close(in);
    }
    if (out != NULL) {
        out->c_close(out);
    }
    if (edge != NULL) {
        edge->c_close(edge);
    }
}


int RdfsBdb::secondaryDbCursor(QString key, DB* db, DBC** cursor, DBT *dbKey, DBT *dbData)
{
    int ret;
    // Init cursor
    if ((ret = db->cursor(db, NULL, cursor, 0)) != 0) {
        return ret;
    }

    // Load key to look up
    QByteArray byteData = key.toUtf8();
    dbKey->data = byteData.data();
    dbKey->size = byteData.size();

    DBC *c = *cursor; // Convenience

    // Get items from cursor and return
    ret = c->c_get(c, dbKey, dbData, DB_SET);

    return ret;
}


int RdfsBdb::checkTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table)
{
    DBC *in_cursor, *out_cursor, *edge_cursor, *join_cursor;

    int flags;
    int ret = findTripleCursor(t, table, &join_cursor, &in_cursor, &out_cursor, &edge_cursor, &flags);
    if (ret != 0) {
        return 0; // error, or empty sub-cursor
    }

    // Init key/data pairs
    DBT key, data;
    memset(&key, 0, sizeof(DBT));
    memset(&data, 0, sizeof(DBT));

    // Count amount of rows and close cursor
    // (Note: c_count only works for current record, and is therefore not usable)
    int count = 0;
    while ((ret = join_cursor->c_get(join_cursor, &key, &data, flags)) == 0) {
        ++count;
    }

    closeCursors(join_cursor, in_cursor, out_cursor, edge_cursor);

    if (ret == DB_NOTFOUND) {
        return count;
    }
    // Other error
    return 0;
}

#ifdef DB_MULTIPLE_RECNO_WRITE_INIT
void RdfsBdb::startBulkLoad()
{
    if (m_bulkLoad) {
        endBulkLoad();
    }

    DB_BTREE_STAT *stats = NULL;
    m_db_primary->stat(m_db_primary, NULL, &stats, DB_FAST_STAT);
    m_nextRecno = stats->bt_nkeys + 1;
    free(stats);

    m_bulkData = (DBT *)malloc(sizeof(DBT));
    memset(m_bulkData, 0, sizeof(DBT));

    const u_int32_t size = 128*1024; // 128 Kb

    m_bulkData->data = malloc(size);
    memset(m_bulkData->data, 0, size);
    m_bulkData->ulen = size;

    DB_MULTIPLE_RECNO_WRITE_INIT(m_bulkPointer, m_bulkData);
    m_bulkLoad = true;
    RdfsLogger::instance()->debug(QString("Started bulk load; first record: %1").arg(m_nextRecno));
}

void RdfsBdb::endBulkLoad()
{
    if (!m_bulkLoad) {
        return;
    }

    RdfsLogger::instance()->debug("Ending bulk load");
    int ret = m_db_primary->put(m_db_primary, NULL, m_bulkData, NULL, DB_MULTIPLE_KEY);
    m_bulkLoad = false;
    free(m_bulkData->data);
    free(m_bulkData);
    m_bulkData = NULL;
    m_bulkPointer = NULL;
}
#endif

void RdfsBdb::insertTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table)
{
    DBT data = tripleToDbt(t);
#ifdef DB_MULTIPLE_RECNO_WRITE_INIT
    if (!m_bulkLoad) {
#endif
        DBT key;
        memset(&key, 0, sizeof(DBT));
        // Do not overwrite dupes - they will remain the same
        int ret = m_db_primary->put(m_db_primary, NULL, &key, &data, DB_APPEND);
#ifdef DB_MULTIPLE_RECNO_WRITE_INIT
    }
    else {
        DB_MULTIPLE_RECNO_WRITE_NEXT(m_bulkPointer, m_bulkData, m_nextRecno, data.data, data.size);
        if (m_bulkPointer == NULL) {
            // Data didn't fit - write records and try again
            RdfsLogger::instance()->debug("Error writing record - trying again!");
            endBulkLoad();
            startBulkLoad();
            DB_MULTIPLE_RECNO_WRITE_NEXT(m_bulkPointer, m_bulkData, m_nextRecno, data.data, data.size);
            if (m_bulkPointer == NULL) {
                // Bulk insert is failing - end bulk load, just insert data
                RdfsLogger::instance()->debug("Bulk failed! falling back to default behavior.");
                endBulkLoad();
                insertTriple(t, table);
            }
        }
        m_nextRecno++;
    }
#endif
    free(data.data);
}

int RdfsBdb::capabilities() const
{
    return (RDFS_CURSOR | RDFS_CACHESTATS);
}

void RdfsBdb::close()
{
    if (m_db_in != NULL) {
        m_db_in->close(m_db_in, 0);
        m_db_in = NULL;
    }
    if (m_db_out != NULL) {
        m_db_out->close(m_db_out, 0);
        m_db_out = NULL;
    }
    if (m_db_edge != NULL) {
        m_db_edge->close(m_db_edge, 0);
        m_db_edge = NULL;
    }
    if (m_db_primary != NULL) {
        m_db_primary->close(m_db_primary, 0);
        m_db_primary = NULL;
    }
    if (m_db_env != NULL) {
        m_db_env->close(m_db_env, 0);
        m_db_env = NULL;
    }
}

bool RdfsBdb::isOpen()
{
    return (m_db_primary != NULL);
}

int RdfsBdb::sizeStringToBytes(const QString& sizeString)
{
    // Get rightmost character(s), define multiplier
    QString rightmost, rightmost2;
    rightmost = sizeString.right(1).toLower();
    rightmost2 = sizeString.right(2).toLower();

    int mult = 1;
    int size = 0;
    bool ok; // Indicator if conversion succeeded

    if (rightmost == "g") {
        mult = 1024 * 1024 * 1024;
        size = sizeString.left(sizeString.size() - 1).toInt(&ok);
    }
    else if (rightmost2 == "gb") {
        mult = 1024 * 1024 * 1024;
        size = sizeString.left(sizeString.size() - 2).toInt(&ok);
    }
    else if (rightmost == "m") {
        mult = 1024 * 1024;
        size = sizeString.left(sizeString.size() - 1).toInt(&ok);
    }
    else if (rightmost2 == "mb") {
        mult = 1024 * 1024;
        size = sizeString.left(sizeString.size() - 2).toInt(&ok);
    }
    else if (rightmost == "k") {
        mult = 1024;
        size = sizeString.left(sizeString.size() - 1).toInt(&ok);
    }
    else if (rightmost2 == "kb") {
        mult = 1024;
        size = sizeString.left(sizeString.size() - 2).toInt(&ok);
    }
    else if (rightmost == "b") {
        mult = 1;
        size = sizeString.left(sizeString.size() - 1).toInt(&ok);
    }
    else {
        mult = 1;
        size = sizeString.toInt(&ok);
    }

    if (ok) {
        return mult * size;
    }
    return -1;
}

void RdfsBdb::setCacheSize(const QString& size)
{
    int bytesize = sizeStringToBytes(size);
    if (bytesize >= 0) {
        setCacheSize(bytesize);
    }
}

void RdfsBdb::setCacheSize(int bytes)
{
    if (m_db_env != 0 && bytes >= 20*1024) {
        u_int32_t gb = bytes / (1024*1024*1024);
        u_int32_t b = bytes % (1024*1024*1024);
        m_db_env->set_cachesize(m_db_env, gb, b, 1);
        m_db_env->set_cache_max(m_db_env, gb, b);
    }
}

void RdfsBdb::setPageSize(const QString& size)
{
    int bytesize = sizeStringToBytes(size);
    if (bytesize >= 0) {
        setPageSize(bytesize);
    }
}

void RdfsBdb::setPageSize(int bytes)
{
    if (!isOpen() && m_db_primary != NULL && bytes >= 512 && bytes <= 64*1024) {
        m_db_primary->set_pagesize(m_db_primary, bytes);
        m_db_primary->set_pagesize(m_db_in, bytes);
        m_db_primary->set_pagesize(m_db_out, bytes);
        m_db_primary->set_pagesize(m_db_edge, bytes);
    }
}

CacheStats RdfsBdb::statistics() const
{
    // Sync dbs
    m_db_primary->sync(m_db_primary, 0);
    m_db_in->sync(m_db_in, 0);
    m_db_out->sync(m_db_out, 0);
    m_db_edge->sync(m_db_edge, 0);

    // Default statistics: all 0s
    CacheStats s = RdfsBackend::statistics();
    // Get per-db statistics
    DB_MPOOL_STAT *globalStats;
    DB_MPOOL_FSTAT **fileStats;

    if (m_db_env->memp_stat(m_db_env, &globalStats, &fileStats, 0) != 0) {
        return s;
    }

    s.cachesize = (globalStats->st_gbytes * 1024*1024*1024) + globalStats->st_bytes;
    if (fileStats != NULL) {
        s.pagesize = (*fileStats)->st_pagesize;
    }
    s.hits = globalStats->st_cache_hit;
    s.misses = globalStats->st_cache_miss;
    s.created = globalStats->st_page_create;
    s.read = globalStats->st_page_in;
    s.written = globalStats->st_page_out;

    free(fileStats);
    free(globalStats);

    // Sum items
    return s;
}


RdfsCursor* RdfsBdb::cursor(Triple q, RdfsTable table)
{
    return new RdfsBdbCursor(this, q, table);
}

