/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef RDFS_RDFSBDB_H
#define RDFS_RDFSBDB_H

#include "rdfsbackend.h"
#include <db.h>

namespace Rdfs {

class RdfsBdb : public Rdfs::RdfsBackend
{
private:
    friend class RdfsBdbCursor;

    enum TripleElement {
        RDFS_SUBJECT = 0,
        RDFS_PREDICATE = 1,
        RDFS_OBJECT = 2,
    };

    DB_ENV *m_db_env; /**< Database environment */
    DB *m_db_primary; /**< Primary database (key = statement) */
    DB *m_db_out;  /**< DB pointer for incoming edges (key = subject) */
    DB *m_db_in;   /**< DB pointer for incoming edges (key = object) */
    DB *m_db_edge; /**< DB pointer for edges (key = predicate) */

#ifdef DB_MULTIPLE_RECNO_WRITE_INIT
    bool m_bulkLoad;
    void *m_bulkPointer;
    DBT *m_bulkData;
    db_recno_t m_nextRecno;
#endif

    DBT tripleToDbt(Rdfs::Triple t);
    Rdfs::Triple dbtToTriple(DBT *data);
    /**
     * Load a secondary key for a data item.
     * @param data The data item
     * @param index The index element that should serve as key
     * @param skey The data item to store the key in.
     */
    static void loadKey(const DBT* data, TripleElement index, DBT *skey);

    /**
     * Start a cursor for a secondary database which will traverse over records with a specific key.
     */
    int secondaryDbCursor(QString key, DB *db, DBC **cursor, DBT *dbKey, DBT *dbData);

    /**
     * Get the cursor to query triples. This will create cursors for the appropriate incoming/outgoing/edge
     * databases and make the main cursor (either a join, or a plain cursor if all elements in the triple or blank).
     * It also sets the flags to traverse the items in the flags variable.
     *
     * Note that normally the @p in @p out and @p edge pointers do not need to be used. These cursors do have to be
     * closed using closeCursors() though.
     *
     * @param t The triple query to find (with optional blank entries)
     * @param table The table the triple is in (currently unsupported)
     * @param main_cursor Address of variable to store pointer to main cursor in
     * @param in Address of variable to store incoming cursor in
     * @param out Address of variable to store outgoing cursor in
     * @param edge Address of variable to store edge cursor in
     * @param flags Address of variable to store traversal flags in.
     * @return 0 on success, other value if for example opening a required cursor failed.
     * @see closeCursors()
     */
    int findTripleCursor(const Rdfs::Triple t, Rdfs::RdfsTable table, DBC** main_cursor, DBC** in, DBC** out, DBC** edge, int* flags);

    /**
     * Close cursors and helper cursors opened by findTripleCursor() earlier.
     */
    void closeCursors(DBC *main_cursor, DBC *in, DBC *out, DBC *edge);

    int sizeStringToBytes(const QString &sizeString);

public:
    RdfsBdb(const QString& dir, const QString& file, const QString& incomingFile, const QString& outgoingFile, const QString& edgeFile,
            const QString& cacheSize = QString(), const QString& pageSize = QString());
    virtual ~RdfsBdb();

    virtual void setPageSize(const QString &size);
    virtual void setPageSize(int bytes);
    virtual void setCacheSize(const QString &size);
    virtual void setCacheSize(int bytes);

    virtual void truncate(Rdfs::RdfsTable table = RDFS_ORIGINAL);
    virtual void removeRule(Rdfs::Rule r);
    virtual QList< Rdfs::Rule > rules();
    virtual void insertRule(Rdfs::Rule r);
    virtual void removeTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table = RDFS_ORIGINAL);
    virtual QList< Rdfs::Triple > findTriples(const Rdfs::Triple& t, Rdfs::RdfsTable table);
    virtual int checkTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table = RDFS_ORIGINAL);
#ifdef DB_MULTIPLE_RECNO_WRITE_INIT
    virtual void startBulkLoad();
    virtual void endBulkLoad();
#endif
    virtual void insertTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table = RDFS_ORIGINAL);
    virtual int capabilities() const;
    virtual void close();
    virtual bool isOpen();
    virtual RdfsCursor* cursor(Triple q, RdfsTable table = RDFS_ORIGINAL);

    // Key extractors
    static int incomingKey(DB *db_handle, const DBT *pkey, const DBT *data, DBT *skey);
    static int outgoingKey(DB *db_handle, const DBT *pkey, const DBT *data, DBT *skey);
    static int edgeKey(DB *db_handle, const DBT *pkey, const DBT *data, DBT *skey);

    virtual CacheStats statistics() const;
};

}

#endif // RDFS_RDFSBDB_H
