/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "rdfsbdbbt.h"

#include <stdlib.h>
#include <QList>
#include <QPair>

#include "triple.h"
#include "rdfsbdbbtcursor.h"
#include "rdfslogger.h"
#include <QFileInfo>
#include <QDir>

using namespace Rdfs;

RdfsBdbBt::RdfsBdbBt(const QString& dir, const QString& file, const QString& updateFile, const QString& rulesFile, const QString& cacheSize, const QString& pageSize):
    RdfsBackend()
{
    m_db_env = NULL;
    for (int i = 0; i < 3; ++i) {
        m_db_original[i] = NULL;
        m_db_update[i] = NULL;
        m_db_rule[i] = NULL;
    }

    if (file.isEmpty() || updateFile.isEmpty() || rulesFile.isEmpty()) {
        return;
    }

    // Remove old cache files. This (a) clears cache, (b) allows opening db with flexible cache sizes
    QDir envdir(dir);
    QFileInfoList cacheFiles = envdir.entryInfoList(QStringList() << "__db.*", QDir::Files | QDir::NoDotAndDotDot);
    foreach(QFileInfo f, cacheFiles) {
        RdfsLogger::instance()->debug(QString("Remove stale cache file: %1").arg(f.absoluteFilePath()));
        QFile::remove(f.absoluteFilePath());
    }

    u_int32_t flags, env_flags; // Database/environment open flags
    int ret; // Return value

    // Create & open environment
    if (db_env_create(&m_db_env, 0) != 0) {
        return;
    }

    if (!cacheSize.isEmpty()) {
        setCacheSize(cacheSize);
    }

    // Open environment
    env_flags = DB_CREATE | DB_INIT_MPOOL;
    if ((ret = m_db_env->open(m_db_env, dir.toUtf8().constData(), env_flags, 0)) != 0) {
        RdfsLogger::instance()->error(QObject::tr("Error opening database environment at %1 (error %2)").arg(dir, QString::number(ret)));
        return;
    }

    for (int i = 0; i < 3; ++i) {
        if (db_create(&(m_db_original[i]), m_db_env, 0) != 0) {
            close();
            return;
        }
        if (db_create(&(m_db_update[i]), m_db_env, 0) != 0) {
            close();
            return;
        }
        if (db_create(&(m_db_rule[i]), m_db_env, 0) != 0) {
            close();
            return;
        }
    }

    if (!pageSize.isEmpty()) {
        setPageSize(pageSize);
    }

    // Create non-existing dbs
    flags = DB_CREATE;

    // Open databases
    QString suffixes[] = { "sop", "pso", "ops" };
    QFileInfo originalFileInfo(file);
    QFileInfo updateFileInfo(updateFile);
    QFileInfo rulesFileInfo(rulesFile);

    for (int i = 0; i < 3; ++i) {
        QString dbFile = originalFileInfo.baseName() + "." + suffixes[i] + "." + originalFileInfo.completeSuffix();
        QString dbName = "original-" + suffixes[i];
        if (m_db_original[i]->open(m_db_original[i], NULL, dbFile.toLatin1().constData(), dbName.toLatin1().constData(), DB_BTREE, flags, 0) != 0) {
            close();
            return;
        }

        dbFile = updateFileInfo.baseName() + "." + suffixes[i] + "." + updateFileInfo.completeSuffix();
        dbName = "update-" + suffixes[i];
        if (m_db_update[i]->open(m_db_update[i], NULL, dbFile.toLatin1().constData(), dbName.toLatin1().constData(), DB_BTREE, flags, 0) != 0) {
            close();
            return;
        }

        dbFile = rulesFileInfo.baseName() + "." + suffixes[i] + "." + rulesFileInfo.completeSuffix();
        dbName = "rules-" + suffixes[i];
        if (m_db_rule[i]->open(m_db_rule[i], NULL, dbFile.toLatin1().constData(), dbName.toLatin1().constData(), DB_BTREE, flags, 0) != 0) {
            close();
            return;
        }
    }

}

RdfsBdbBt::~RdfsBdbBt()
{
    if (isOpen()) {
        close();
    }
}

DBT RdfsBdbBt::tripleToDbt(Triple t, RdfsBdbBt::INDEX_ORDER order)
{
    QByteArray strings[3];
    strings[0] = t.subject.toUtf8();
    strings[1] = t.predicate.toUtf8();
    strings[2] = t.object.toUtf8();

    quint16 lengths[3];
    int length = 0;
    for (int i = 0; i < 3; ++i) {
        lengths[i] = strings[i].length();
        length += strings[i].length();
    }
    int dataLength = length*sizeof(char) + 3*sizeof(quint16); // 3 length indicators (int16)

    char *data = (char *)malloc(dataLength);
    memset(data, 0, dataLength);

    // Index order
    int indexes[3];

    switch(order) {
        default:
        case SOP:
            indexes[0] = 0;
            indexes[1] = 2;
            indexes[2] = 1;
            break;
        case PSO:
            indexes[0] = 1;
            indexes[1] = 0;
            indexes[2] = 2;
            break;
        case OPS:
            indexes[0] = 2;
            indexes[1] = 1;
            indexes[2] = 0;
            break;
    }

    // Write data
    for (int i = 0; i < 3; ++i) {
        int index = indexes[i];
        // Length
        memcpy(data, &lengths[index], sizeof(quint16));
        data += sizeof(quint16);
        // Data
        memcpy(data, strings[index], lengths[index] * sizeof(char));
        data += lengths[index] * sizeof(char);
    }

    // Reset pointer to beginning
    data -= dataLength;

    // Construct result
    DBT result;
    memset(&result, 0, sizeof(DBT));
    result.data = data;
    result.size = dataLength;

    return result;
}

Triple RdfsBdbBt::dbtToTriple(DBT* data, RdfsBdbBt::INDEX_ORDER order)
{
    Triple t;

    QString strings[3];

    // Decode data: (length + string)^3
    char *p = (char *)data->data;
    quint16 length = 0;
    for (int i = 0; i < 3; ++i) {
        memcpy(&length, p, sizeof(quint16));
        p += sizeof(quint16);
        if (length > 0) {
            strings[i] = QString::fromUtf8(p, length);
        }
        else {
            strings[i] = QString();
        }
        p += length;
    }

    // Build triple
    switch(order) {
        default:
        case SOP:
            t.subject = strings[0];
            t.predicate = strings[2];
            t.object = strings[1];
            break;
        case PSO:
            t.subject = strings[1];
            t.predicate = strings[0];
            t.object = strings[2];
            break;
        case OPS:
            t.subject = strings[2];
            t.predicate = strings[1];
            t.object = strings[0];
            break;
    }
    return t;
}

DB** RdfsBdbBt::dbs(RdfsTable table)
{
    switch(table) {
        case RDFS_ORIGINAL:
        default:
            return m_db_original;
        case RDFS_UPDATE:
            return m_db_update;
        case RDFS_RULE:
            return m_db_rule;
    }
}

void RdfsBdbBt::truncate(Rdfs::RdfsTable table)
{
    DB** db = dbs(table);
    for (int i = 0; i < 3; ++i) {
        db[i]->truncate(db[i], NULL, NULL, 0);
    }
}

void RdfsBdbBt::removeRule(Rdfs::Rule r)
{
    // TODO Stub
}

QList< QPair< Rdfs::Triple, Rdfs::Triple > > RdfsBdbBt::rules()
{
    // TODO Stub
    return QList< QPair< Rdfs::Triple, Rdfs::Triple > >();
}

void RdfsBdbBt::insertRule(Rdfs::Rule r)
{
    // TODO Stub
}

void RdfsBdbBt::removeTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table)
{
    RdfsBdbBtCursor *c = new RdfsBdbBtCursor(this, t, table, dbs(table));
    if (c->open()) {
        c->removeCurrent();
    }
    c->close();
    delete c;
}

QList< Rdfs::Triple > RdfsBdbBt::findTriples(const Rdfs::Triple& t, Rdfs::RdfsTable table)
{
    QList<Triple> result;
    RdfsBdbBtCursor *c = new RdfsBdbBtCursor(this, t, table, dbs(table));
    if (c->open()) {
        while (!c->eof()) {
            result << c->current();
            c->next();
        }
        c->close();
    }
    delete c;
    return result;
}

int RdfsBdbBt::checkTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table)
{
    int result = 0;
    RdfsBdbBtCursor *c = new RdfsBdbBtCursor(this, t, table, dbs(table));
    if (c->open()) {
        while (!c->eof()) {
            ++result;
            c->next();
        }
        c->close();
    }
    delete c;

    return result;
}

void RdfsBdbBt::insertTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table)
{
    DB** db = dbs(table);
    INDEX_ORDER orders[] = { SOP, PSO, OPS };
    // Add item to all three indexes
    for (int i = 0; i < 3; ++i) {
        // Create key - data is empty
        DBT key, data;
        key = tripleToDbt(t, orders[i]);
        memset(&data, 0, sizeof(DBT));
        // And insert it
        db[orders[i]]->put(db[orders[i]], NULL, &key, &data, 0);
        free(key.data);
    }
}

int RdfsBdbBt::capabilities() const
{
    return (RDFS_CURSOR | RDFS_CACHESTATS);
}

void RdfsBdbBt::close()
{
    for (int i = 0; i < 3; ++i) {
        if (m_db_original[i] != NULL) {
            m_db_original[i]->close(m_db_original[i], 0);
            m_db_original[i] = NULL;
        }
        if (m_db_update[i] != NULL) {
            m_db_update[i]->close(m_db_update[i], 0);
            m_db_update[i] = NULL;
        }
        if (m_db_rule[i] != NULL) {
            m_db_rule[i]->close(m_db_rule[i], 0);
            m_db_rule[i] = NULL;
        }
    }

    if (m_db_env != NULL) {
        m_db_env->close(m_db_env, 0);
        m_db_env = NULL;
    }
}

bool RdfsBdbBt::isOpen()
{
    return (m_db_original[0] != NULL);
}

void RdfsBdbBt::startBulkLoad()
{
    Rdfs::RdfsBackend::startBulkLoad();
}

void RdfsBdbBt::endBulkLoad()
{
    Rdfs::RdfsBackend::endBulkLoad();
}

bool RdfsBdbBt::hasTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table)
{
    bool result;
    RdfsCursor *c = new RdfsBdbBtCursor(this, t, table, dbs(table));
    if (c->open()) {
        result = true;
        c->close();
    }
    else {
        result = false;
    }
    delete c;
    return result;
}

Rdfs::RdfsCursor* RdfsBdbBt::cursor(Rdfs::Triple q, Rdfs::RdfsTable table)
{
    return new RdfsBdbBtCursor(this, q, table, dbs(table));
}

int RdfsBdbBt::sizeStringToBytes(const QString& sizeString)
{
    // Get rightmost character(s), define multiplier
    QString rightmost, rightmost2;
    rightmost = sizeString.right(1).toLower();
    rightmost2 = sizeString.right(2).toLower();

    int mult = 1;
    int size = 0;
    bool ok; // Indicator if conversion succeeded

    if (rightmost == "g") {
        mult = 1024 * 1024 * 1024;
        size = sizeString.left(sizeString.size() - 1).toInt(&ok);
    }
    else if (rightmost2 == "gb") {
        mult = 1024 * 1024 * 1024;
        size = sizeString.left(sizeString.size() - 2).toInt(&ok);
    }
    else if (rightmost == "m") {
        mult = 1024 * 1024;
        size = sizeString.left(sizeString.size() - 1).toInt(&ok);
    }
    else if (rightmost2 == "mb") {
        mult = 1024 * 1024;
        size = sizeString.left(sizeString.size() - 2).toInt(&ok);
    }
    else if (rightmost == "k") {
        mult = 1024;
        size = sizeString.left(sizeString.size() - 1).toInt(&ok);
    }
    else if (rightmost2 == "kb") {
        mult = 1024;
        size = sizeString.left(sizeString.size() - 2).toInt(&ok);
    }
    else if (rightmost == "b") {
        mult = 1;
        size = sizeString.left(sizeString.size() - 1).toInt(&ok);
    }
    else {
        mult = 1;
        size = sizeString.toInt(&ok);
    }

    if (ok) {
        return mult * size;
    }
    return -1;
}

void RdfsBdbBt::setCacheSize(const QString& size)
{
    int bytesize = sizeStringToBytes(size);
    if (bytesize >= 0) {
        setCacheSize(bytesize);
    }
}

void RdfsBdbBt::setCacheSize(int bytes)
{
    if (m_db_env != 0 && bytes >= 20*1024) {
        u_int32_t gb = bytes / (1024*1024*1024);
        u_int32_t b = bytes % (1024*1024*1024);
        m_db_env->set_cachesize(m_db_env, gb, b, 1);
        m_db_env->set_cache_max(m_db_env, gb, b);
    }
}

void RdfsBdbBt::setPageSize(const QString& size)
{
    int bytesize = sizeStringToBytes(size);
    if (bytesize >= 0) {
        setPageSize(bytesize);
    }
}

void RdfsBdbBt::setPageSize(int bytes)
{
    if (!isOpen() && m_db_original[0] != NULL && bytes >= 512 && bytes <= 64*1024) {
        for (int i = 0; i < 3; ++i) {
            m_db_original[i]->set_pagesize(m_db_original[i], bytes);
            m_db_update[i]->set_pagesize(m_db_update[i], bytes);
            m_db_rule[i]->set_pagesize(m_db_rule[i], bytes);
        }
    }
}

CacheStats RdfsBdbBt::statistics() const
{
    // Sync dbs
    for (int i = 0; i < 3; ++i) {
        m_db_original[i]->sync(m_db_original[i], 0);
        m_db_update[i]->sync(m_db_update[i], 0);
        m_db_rule[i]->sync(m_db_rule[i], 0);
    }

    // Default statistics: all 0s
    CacheStats s = RdfsBackend::statistics();
    // Get per-db statistics
    DB_MPOOL_STAT *globalStats;
    DB_MPOOL_FSTAT **fileStats;

    if (m_db_env->memp_stat(m_db_env, &globalStats, &fileStats, 0) != 0) {
        return s;
    }

    s.cachesize = (globalStats->st_gbytes * 1024*1024*1024) + globalStats->st_bytes;
    if (fileStats != NULL) {
        s.pagesize = (*fileStats)->st_pagesize;
    }
    s.hits = globalStats->st_cache_hit;
    s.misses = globalStats->st_cache_miss;
    s.created = globalStats->st_page_create;
    s.read = globalStats->st_page_in;
    s.written = globalStats->st_page_out;

    free(fileStats);
    free(globalStats);

    // Sum items
    return s;
}
