/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "rdfsbdbtriplet.h"

#include <stdlib.h>
#include <QSet>
#include <QDir>
#include <QTextStream>

#include "rdfslogger.h"
#include "bdbtripletbtreecursor.h"

using namespace Rdfs;

RdfsBdbTripleT::RdfsBdbTripleT(const QString& dir, const QString& file, const QString& updateFile, const QString& rulesFile, const QString& cacheSize, const QString& pageSize)
    // Note: use QString(dir) as QString::replace() is non-const
    : RdfsTripleTBackend("bdbtt-" + QString(dir).replace("/", "-").replace("\\", "-"))
{
    m_db_env = NULL;
    m_db_original = NULL;
    m_db_update = NULL;
    m_db_rules = NULL;

    m_dir = dir;
    m_file = file;
    m_updateFile = updateFile;
    m_rulesFile = rulesFile;

    if (file.isEmpty() || updateFile.isEmpty() || rulesFile.isEmpty()) {
        return;
    }

    // Remove old cache files. This (a) clears cache, (b) allows opening db with flexible cache sizes
    QDir envdir(dir);
    QFileInfoList cacheFiles = envdir.entryInfoList(QStringList() << "__db.*", QDir::Files | QDir::NoDotAndDotDot);
    foreach(QFileInfo f, cacheFiles) {
        RdfsLogger::instance()->debug(QString("Remove stale cache file: %1").arg(f.absoluteFilePath()));
        QFile::remove(f.absoluteFilePath());
    }

    // Start opening DB environment
    u_int32_t flags, env_flags; // Database/environment open flags
    int ret; // Return value

    // Create & open environment
    if (db_env_create(&m_db_env, 0) != 0) {
        return;
    }

    if (!cacheSize.isEmpty()) {
        setCacheSize(cacheSize);
    }

    env_flags = DB_CREATE | DB_INIT_MPOOL;
    if ((ret = m_db_env->open(m_db_env, dir.toUtf8().constData(), env_flags, 0)) != 0) {
        RdfsLogger::instance()->error(QObject::tr("Error opening database environment at %1 (error %2)").arg(dir, QString::number(ret)));
        return;
    }

    if (db_create(&m_db_original, m_db_env, 0) != 0) {
        close();
        return;
    }
    if (db_create(&m_db_update, m_db_env, 0) != 0) {
        close();
        return;
    }
    if (db_create(&m_db_rules, m_db_env, 0) != 0) {
        close();
        return;
    }


    if (!pageSize.isEmpty()) {
        setPageSize(pageSize);
    }

    // Create non-existing dbs
    flags = DB_CREATE;

    // Open databases
    if (m_db_original->open(m_db_original, NULL, file.toLatin1().constData(), "original", DB_HASH, flags, 0) != 0) {
        close();
        return;
    }
    if (m_db_update->open(m_db_update, NULL, updateFile.toLatin1().constData(), "update", DB_HASH, flags, 0) != 0) {
        close();
        return;
    }
    if (m_db_rules->open(m_db_rules, NULL, rulesFile.toLatin1().constData(), "rules", DB_HASH, flags, 0) != 0) {
        close();
        return;
    }

    // Load B-trees file
    QString base = m_dir + QDir::separator() + file;
    QFileInfo baseInfo(base);
    base = baseInfo.absolutePath() + QDir::separator() + baseInfo.baseName();
    QFileInfo btreesInfo(base + "-btrees.txt");
    if (btreesInfo.exists() && btreesInfo.isReadable() && btreesInfo.isFile()) {
        QFile bt(btreesInfo.absoluteFilePath());
        bt.open(QIODevice::Text | QIODevice::ReadOnly);
        QTextStream stream(&bt);

        while (!stream.atEnd()) {
            QString db = stream.readLine().trimmed();
            RdfsLogger::instance()->debug("Opening B-tree database: " + db);
            if (btreeDatabase(db, RDFS_ORIGINAL, false) == NULL) {
                bt.close();
                close();
                return;
            }
        }
        bt.close();
    }

}

RdfsBdbTripleT::~RdfsBdbTripleT()
{
    if (isOpen()) {
        close();
    }
}

DB* RdfsBdbTripleT::database(RdfsTable table)
{
    switch(table) {
        default:
        case RDFS_ORIGINAL:
            return m_db_original;
        case RDFS_UPDATE:
            return m_db_update;
        case RDFS_RULE:
            return m_db_rules;
    }
}

DB* RdfsBdbTripleT::btreeDatabase(const QString& atom, const Rdfs::RdfsRole& role, const Rdfs::RdfsTable& table, bool updateBtreeIndex)
{
    QString id;
    switch(role) {
        case RDFS_SUBJECT:
            id = atom + "-s";
            break;
        case RDFS_PREDICATE:
            id = atom + "-p";
            break;
        case RDFS_OBJECT:
            id = atom + "-o";
            break;
    }

    return btreeDatabase(id, table, updateBtreeIndex);
}

DB* RdfsBdbTripleT::btreeDatabase(const QString& id, const Rdfs::RdfsTable& table, bool updateBtreeIndex)
{
    QMap<QString, DB *> *container;
    QString tablebase, basefile;
    switch(table) {
        case RDFS_ORIGINAL:
        default:
            container = &m_btreesOriginal;
            tablebase = QFileInfo(m_file).baseName();
            basefile = m_file;
            break;
        case RDFS_UPDATE:
            container = &m_btreesUpdate;
            tablebase = QFileInfo(m_updateFile).baseName();
            basefile = m_updateFile;
            break;
    }

    // Database may already be loaded
    if (container->contains(id)) {
        return container->value(id);
    }

    // Create database
    QString fullDbName(tablebase + "-" + id);
    DB *btree;
    if (db_create(&btree, m_db_env, 0) != 0) {
        return NULL;
    }

    // Create non-existing dbs
    u_int32_t flags = DB_CREATE;

    // Open databases
    if (btree->open(btree, NULL, fullDbName.toLatin1().constData(), fullDbName.toLatin1().constData(), DB_BTREE, flags, 0) != 0) {
        btree->close(btree, 0);
        return NULL;
    }

    container->insert(id, btree);

    if (updateBtreeIndex) {
        QString base = m_dir + QDir::separator() + m_file;
        QFileInfo baseInfo(base);
        QFileInfo pathInfo(baseInfo.path());
        base = baseInfo.absolutePath() + QDir::separator() + baseInfo.baseName();
        QFileInfo btreesInfo(base + "-btrees.txt");
        if ((btreesInfo.exists() && btreesInfo.isWritable() && btreesInfo.isFile()) ||
            (!btreesInfo.exists() && pathInfo.isWritable() && pathInfo.isDir())) {
            QFile bti(btreesInfo.absoluteFilePath());
            bti.open(QIODevice::Append | QIODevice::WriteOnly | QIODevice::Text);
            QTextStream stream(&bti);
            stream << id << "\n";
            bti.close();
        }
    }

    return btree;
}


QStringList RdfsBdbTripleT::atoms(Rdfs::RdfsTable table)
{
    DBC *cursor = NULL;

    DB *db = database(table);
    int ret = db->cursor(db, NULL, &cursor, 0);

    DBT key, data;
    memset(&key, 0, sizeof(DBT));
    memset(&data, 0, sizeof(DBT));

    QSet<QString> result;

    while ((ret = cursor->c_get(cursor, &key, &data, DB_NEXT)) == 0) {
        QString atom = QString::fromUtf8((const char*)key.data, key.size);
        atom = atom.remove(0, 2); // Remove role prefix
        result.insert(atom);
    }
    if (cursor != NULL) {
        cursor->c_close(cursor);
    }

    return result.toList();
}

void RdfsBdbTripleT::putPayload(const QString &atom, const RdfsRole &role, QByteArray data, Rdfs::RdfsTable table)
{
    DBT key, payload;
    memset(&key, 0, sizeof(DBT));
    memset(&payload, 0, sizeof(DBT));

    QString prefix;
    if (role == RDFS_SUBJECT) {
        prefix = "s:";
    }
    else if (role == RDFS_PREDICATE) {
        prefix = "p:";
    }
    else {
        prefix = "o:";
    }

    RdfsLogger::instance()->debug("Putting: " + prefix + atom);

    QByteArray utf8key = (prefix + atom).toUtf8();
    key.data = utf8key.data();
    key.size = utf8key.size();

    payload.data = data.data();
    payload.size = data.size();

    DB *db = database(table);
    db->put(db, NULL, &key, &payload, 0);
}

QByteArray RdfsBdbTripleT::payload(const QString &atom, const RdfsRole &role, Rdfs::RdfsTable table)
{
    DBT key, payload;
    memset(&key, 0, sizeof(DBT));
    memset(&payload, 0, sizeof(DBT));

    QString prefix;
    if (role == RDFS_SUBJECT) {
        prefix = "s:";
    }
    else if (role == RDFS_PREDICATE) {
        prefix = "p:";
    }
    else {
        prefix = "o:";
    }

    RdfsLogger::instance()->debug("Getting: " + prefix + atom);

    QByteArray utf8key = (prefix + atom).toUtf8();
    key.data = utf8key.data();
    key.size = utf8key.size();

    DB *db = database(table);
    db->get(db, NULL, &key, &payload, 0);

    return QByteArray((char *)payload.data, payload.size);
}

DBT RdfsBdbTripleT::btreeDbt(const QString& secondary, const QString& tertiary)
{
    QByteArray strings[2];
    strings[0] = secondary.toUtf8();
    strings[1] = tertiary.toUtf8();

    quint16 lengths[2];
    int length = 0;
    for (int i = 0; i < 2; ++i) {
        lengths[i] = strings[i].length();
        length += strings[i].length();
    }
    int dataLength = length*sizeof(char) + 2*sizeof(quint16); // 3 length indicators (int16)

    char *data = (char *)malloc(dataLength);
    memset(data, 0, dataLength);

    for (int i = 0; i < 2; ++i) {
        // Length
        memcpy(data, &lengths[i], sizeof(quint16));
        data += sizeof(quint16);
        // Data
        memcpy(data, strings[i], lengths[i] * sizeof(char));
        data += lengths[i] * sizeof(char);
    }

    // Reset pointer to beginning
    data -= dataLength;

    // Construct result
    DBT result;
    memset(&result, 0, sizeof(DBT));
    result.data = data;
    result.size = dataLength;

    return result;
}

Triple RdfsBdbTripleT::btreeTripleFromDbt(DBT* record, DBT* data, const QString& primary)
{
    Triple ordered, t;

    QString strings[2];

    // Decode data: (length + string)^3
    char *p = (char *)record->data;
    quint16 length = 0;
    for (int i = 0; i < 2; ++i) {
        memcpy(&length, p, sizeof(quint16));
        p += sizeof(quint16);
        if (length > 0) {
            strings[i] = QString::fromUtf8(p, length);
        }
        else {
            strings[i] = QString();
        }
        p += length;
    }

    // No need to re-order items; base will do it
    t.subject = primary;
    t.predicate = strings[0];
    t.object = strings[1];
    t.distance = 0;

    // Get distance from data DBT
    if (data->size) {
        memcpy(&t.distance, data->data, qMin(sizeof(t.distance), (long unsigned int)data->size));
    }
    // Else no distance data: distance was 0

    return t;
}

void RdfsBdbTripleT::insertBtreeTriple(const QString& atom, const Rdfs::RdfsRole& role, const QString& secondary, const QString& tertiary, unsigned char distance, RdfsTable table)
{
    DB *btree = btreeDatabase(atom, role, table);
    DBT key = btreeDbt(secondary, tertiary);
    DBT data;
    memset(&data, 0, sizeof(DBT));

    if (distance) {
        data.data = &distance;
        data.size = sizeof(distance);

        // If triple already exists, make sure the new one has a smaller distance
        DBT oldData;
        memset(&oldData, 0, sizeof(DBT));
        int ret = btree->get(btree, NULL, &key, &oldData, 0);
        if (ret == 0) {
            // Item was found
            unsigned char oldDistance = 0;
            if (oldData.size) { // If no size, i.e. no data, assume distance 0
                memcpy(&oldDistance, oldData.data, sizeof(oldDistance));
            }

            if (oldDistance <= distance) {
                // Ignore new triple: free resources and exit
                free(key.data);
                return;
            }
        }
    }

    // And insert it (or overwrite previous)
    btree->put(btree, NULL, &key, &data, 0);
    free(key.data);
}

void RdfsBdbTripleT::removeBtreeTriple(const QString& atom, const Rdfs::RdfsRole& role, const QString& secondary, const QString& tertiary, RdfsTable table)
{
    DB *btree = btreeDatabase(atom, role, table);
    // Note: triple is always complete
    DBT key = btreeDbt(secondary, tertiary);

    btree->del(btree, NULL, &key, 0);
    free(key.data);
}

RdfsCursor* RdfsBdbTripleT::btreeCursor(const QString& atom, const Rdfs::RdfsRole& role, const QString& secondary, const QString& tertiary, RdfsTable table)
{
    Triple q;
    q.subject = atom;
    q.predicate = secondary;
    q.object = tertiary;

    return new BdbTripleTBtreeCursor(this, q, role, table);
}


void RdfsBdbTripleT::truncate(Rdfs::RdfsTable table)
{
    database(table)->truncate(database(table), NULL, NULL, 0);
}

int RdfsBdbTripleT::capabilities() const
{
    int base = RdfsTripleTBackend::capabilities();
    return (base | RDFS_CACHESTATS);
}

void RdfsBdbTripleT::close()
{
    if (m_db_original != NULL) {
        m_db_original->close(m_db_original, 0);
        m_db_original = NULL;
    }
    if (m_db_update != NULL) {
        m_db_update->close(m_db_update, 0);
        m_db_update = NULL;
    }
    if (m_db_rules != NULL) {
        m_db_rules->close(m_db_rules, 0);
        m_db_rules = NULL;
    }

    QMap<QString, DB*>::const_iterator i;
    for (i = m_btreesOriginal.constBegin(); i != m_btreesOriginal.constEnd(); ++i) {
        DB *btreedb = i.value();
        btreedb->close(btreedb, 0);
    }
    for (i = m_btreesUpdate.constBegin(); i != m_btreesUpdate.constEnd(); ++i) {
        DB *btreedb = i.value();
        btreedb->close(btreedb, 0);
    }

    if (m_db_env != NULL) {
        m_db_env->close(m_db_env, 0);
        m_db_env = NULL;
    }
}

bool RdfsBdbTripleT::isOpen()
{
    return (m_db_original != NULL);
}

int RdfsBdbTripleT::sizeStringToBytes(const QString& sizeString)
{
    // Get rightmost character(s), define multiplier
    QString rightmost, rightmost2;
    rightmost = sizeString.right(1).toLower();
    rightmost2 = sizeString.right(2).toLower();

    int mult = 1;
    int size = 0;
    bool ok; // Indicator if conversion succeeded

    if (rightmost == "g") {
        mult = 1024 * 1024 * 1024;
        size = sizeString.left(sizeString.size() - 1).toInt(&ok);
    }
    else if (rightmost2 == "gb") {
        mult = 1024 * 1024 * 1024;
        size = sizeString.left(sizeString.size() - 2).toInt(&ok);
    }
    else if (rightmost == "m") {
        mult = 1024 * 1024;
        size = sizeString.left(sizeString.size() - 1).toInt(&ok);
    }
    else if (rightmost2 == "mb") {
        mult = 1024 * 1024;
        size = sizeString.left(sizeString.size() - 2).toInt(&ok);
    }
    else if (rightmost == "k") {
        mult = 1024;
        size = sizeString.left(sizeString.size() - 1).toInt(&ok);
    }
    else if (rightmost2 == "kb") {
        mult = 1024;
        size = sizeString.left(sizeString.size() - 2).toInt(&ok);
    }
    else if (rightmost == "b") {
        mult = 1;
        size = sizeString.left(sizeString.size() - 1).toInt(&ok);
    }
    else {
        mult = 1;
        size = sizeString.toInt(&ok);
    }

    if (ok) {
        return mult * size;
    }
    return -1;
}

void RdfsBdbTripleT::setCacheSize(const QString& size)
{
    int bytesize = sizeStringToBytes(size);
    if (bytesize >= 0) {
        setCacheSize(bytesize);
    }
}

void RdfsBdbTripleT::setCacheSize(int bytes)
{
    if (m_db_env != 0 && bytes >= 20*1024) {
        u_int32_t gb = bytes / (1024*1024*1024);
        u_int32_t b = bytes % (1024*1024*1024);
        m_db_env->set_cachesize(m_db_env, gb, b, 1);
        m_db_env->set_cache_max(m_db_env, gb, b);
    }
}

void RdfsBdbTripleT::setPageSize(const QString& size)
{
    int bytesize = sizeStringToBytes(size);
    if (bytesize >= 0) {
        setPageSize(bytesize);
    }
}

void RdfsBdbTripleT::setPageSize(int bytes)
{
    if (!isOpen() && m_db_original != NULL && bytes >= 512 && bytes <= 64*1024) {
        m_db_original->set_pagesize(m_db_original, bytes);
        m_db_update->set_pagesize(m_db_original, bytes);
        m_db_rules->set_pagesize(m_db_rules, bytes);
        RdfsTripleTBackend::setPageSize(bytes);
    }
}

Rdfs::CacheStats RdfsBdbTripleT::statistics() const
{
    // Sync dbs
    m_db_original->sync(m_db_original, 0);
    m_db_update->sync(m_db_update, 0);
    m_db_rules->sync(m_db_rules, 0);

    // Default statistics: all 0s
    CacheStats s = RdfsBackend::statistics();
    // Get per-db statistics
    DB_MPOOL_STAT *globalStats;
    DB_MPOOL_FSTAT **fileStats;

    if (m_db_env->memp_stat(m_db_env, &globalStats, &fileStats, 0) != 0) {
        return s;
    }

    s.cachesize = (globalStats->st_gbytes * 1024*1024*1024) + globalStats->st_bytes;
    if (fileStats != NULL) {
        s.pagesize = (*fileStats)->st_pagesize;
    }
    s.hits = globalStats->st_cache_hit;
    s.misses = globalStats->st_cache_miss;
    s.created = globalStats->st_page_create;
    s.read = globalStats->st_page_in;
    s.written = globalStats->st_page_out;

    free(fileStats);
    free(globalStats);

    // Sum items
    return s;
}

