/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef RDFS_RDFSCURSOR_H
#define RDFS_RDFSCURSOR_H

#include "triple.h"
#include "rdfsbackend.h"

namespace Rdfs {

class RdfsCursor
{
private:
    Triple m_query;
    RdfsTable m_table;
    RdfsBackend *m_backend;

protected:
    bool m_open;

public:
    /**
     * Construct the cursor based on a query. This sets the query, but does not
     * open the cursor yet. Use open() to open the cursor.
     * @param backend The data store (backend) used to run the query
     * @param query The query triple (with blank atoms to be filled in) of the cursor.
     * @param table The table the query should be performed in
     * @see open()
     */
    RdfsCursor(Rdfs::RdfsBackend* backend, const Rdfs::Triple& query, Rdfs::RdfsTable table = RDFS_ORIGINAL);

    /**
     * Destructor.
     *
     * Note that if you reimplement this class, you close any open cursors first!
     */
    virtual ~RdfsCursor();

    /**
     * Get the query triple for this cursor.
     */
    Triple queryTriple() const { return m_query; };

    /**
     * Get the query table for this cursor.
     */
    RdfsTable queryTable() const { return m_table; };

    /**
     * Get the triple at the current position.
     * Note: if the cursor is at the end, this returns an empty triple.
     * Check eof() or the return value of next() to see if a triple is available.
     */
    virtual Triple current() const = 0;

    /**
     * Move to the next item in the cursor. The cursor will not move beyond the
     * end. Note that if the cursor is at the end (after the move), getting the
     * triple (current()) will return an empty triple.
     *
     * @return true if a new triple is available, or false if the cursor is not open
     *         or the cursor is at the end.
     */
    virtual bool next() = 0;

    /**
     * Reset the cursor to the initial position.
     *
     * The default implementation closes and reopens the cursor. If you reimplement
     * this class, override reset() to something more efficient if your backend
     * supports it.
     */
    virtual void reset() { close(); open(); };

    /**
     * Open the cursor for the set query.
     *
     * If the cursor can be opened successfully and there are results for the
     * query, this will return true. If opening fails for some reason or if there
     * are no results for the query, then this returns false.
     *
     * @return true if opening the cursor is successful and there are results,
     *         false otherwise.
     */
    virtual bool open() = 0;

    /**
     * Close the cursor. Getting new triples or moving the cursor will not do
     * anything anymoer.
     */
    virtual void close() = 0;

    /**
     * Check if the cursor is at the end. If this returns true, there is no
     * current triple. This also returns true if the cursor is closed.
     */
    virtual bool eof() const = 0;
};

}

#endif // RDFS_RDFSCURSOR_H
