/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "rdfsmysql.h"

#include <QSqlRecord>
#include <QSqlQuery>
#include <QSqlField>
#include <QStringList>

#include <QtDebug>

using namespace Rdfs;


RdfsMySql::RdfsMySql(const QString& databaseName, const QString& hostname, int port, const QString& username, const QString& password, const QString &socket)
    : RdfsSql("QMYSQL")
{
    // Due to virtuals do not call RdfsSql constructor with all URL parts,
    // otherwise initializeTables() will not use RdfsMySql implementation

    qDebug() << "Db: " << databaseName << " / user: " << username << " / pass: " << password;
    setDatabaseName(databaseName);
    setUserName(username);
    setPassword(password);

    if (!socket.isEmpty()) {
        qDebug() << "Socket: " << socket;
        setConnectOptions("UNIX_SOCKET=" + socket);
    }
    else {
        qDebug() << "Host: " << hostname << ":" << QString::number(port);
        setHostName(hostname);
        setPort(port);
    }

    if (open()) {
        initializeTables();
    }
}

RdfsMySql::~RdfsMySql()
{

}

void RdfsMySql::initializeRdfsOriginal()
{
    if (!databaseTables().contains("rdfs_original")) {
        QSqlQuery query(*this);
        // Note: triple cannot be forced to be unique by index; index becomes too large for MySQL (max 1000)
        query.prepare("CREATE TABLE rdfs_original ("
            "id        INTEGER NOT NULL AUTO_INCREMENT, "
            "subject   VARCHAR(512) NOT NULL, "
            "predicate VARCHAR(512) NOT NULL, "
            "object    VARCHAR(512) NOT NULL, "
            "PRIMARY KEY (id), "
            "INDEX idx_subject   (subject), "
            "INDEX idx_predicate (predicate), "
            "INDEX idx_object    (object))"
        );
        executeQuery(&query);
    }
}

void RdfsMySql::initializeRdfsRules() {
    if (!databaseTables().contains("rdfs_rule_triple")) {
        QSqlQuery query(*this);
        // Note: triple cannot be forced to be unique by index; index becomes too large for MySQL (max 1000)
        query.prepare("CREATE TABLE rdfs_rule_triple ("
            "id        INTEGER NOT NULL AUTO_INCREMENT, "
            "subject   VARCHAR(512) NOT NULL, "
            "predicate VARCHAR(512) NOT NULL, "
            "object    VARCHAR(512) NOT NULL, "
            "PRIMARY KEY (id), "
            "INDEX idx_subject   (subject), "
            "INDEX idx_predicate (predicate), "
            "INDEX idx_object    (object))"
            );
        executeQuery(&query);
    }

    if (!databaseTables().contains("rdfs_rule")) {
        QSqlQuery query(*this);
        query.prepare("CREATE TABLE rdfs_rule ("
            "source    INTEGER NOT NULL, "
            "target    INTEGER NOT NULL, "
            "PRIMARY KEY (source, target), "
            "INDEX idx_source    (source), "
            "INDEX idx_target    (target))");
        executeQuery(&query);
    }
}

void RdfsMySql::initializeRdfsUpdate()
{
    if (!databaseTables().contains("rdfs_update")) {
        QSqlQuery query(*this);
        // Note: triple cannot be forced to be unique by index; index becomes too large for MySQL (max 1000)
        query.prepare("CREATE TABLE rdfs_update ("
            "id        INTEGER NOT NULL AUTO_INCREMENT, "
            "subject   VARCHAR(512) NOT NULL, "
            "predicate VARCHAR(512) NOT NULL, "
            "object    VARCHAR(512) NOT NULL, "
            "PRIMARY KEY (id), "
            "INDEX idx_subject   (subject), "
            "INDEX idx_predicate (predicate), "
            "INDEX idx_object    (object))"
        );
        executeQuery(&query);
    }
}


QStringList RdfsMySql::databaseTables()
{
    QSqlQuery query(*this);
    QStringList tables;
    query.prepare("SHOW TABLES");
    executeQuery(&query);
    while (query.next()) {
        tables << query.record().value(0).toString();
    }
    return tables;
}

void RdfsMySql::insertTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table)
{
    // Triple has to be valid
    if (t.subject.isEmpty() || t.predicate.isEmpty() || t.object.isEmpty()) {
        return;
    }

    // Variables
    QString tableStr = tableName(table);

    // Start transaction
    QSqlQuery query(*this);

    // Insert record
    query.prepare("INSERT INTO " + tableStr + " (subject, predicate, object) "
                  "VALUES (?, ?, ?)");
    query.bindValue(0, t.subject);
    query.bindValue(1, t.predicate);
    query.bindValue(2, t.object);
    executeQuery(&query);
}

