/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "subpropertyrule.h"

#include <QSet>
#include <QStringList>
#include <QQueue>

using namespace Rdfs;

SubpropertyRule::SubpropertyRule()
 : TransitiveRule()
{

}

bool SubpropertyRule::check(Rdfs::Triple t, Rdfs::RdfsTable table, RdfsBackend *store)
{
    if (!containsMap(RDFS_SP)) {
        return false;
    }

    // We always want to check step 4 below, if the refl/trans check fails
    bool tInEntailment = false;

    if (t.predicate == map(RDFS_SP)) {
        // Reflexivity first; if A == B, then this will also catch the transitivity.
        // If this isn't found, then the transitivity will also find nothing
        if (t.subject == t.object) {
            tInEntailment = checkReflexive(t, table, store);
        }
        else {
            // Check transitivity
            tInEntailment = checkTransitive(t, table, store);
        }
    }

    if (!tInEntailment) {
        // Check rule 2b/step 4; note that the predicate may be in rhodf in this
        // implementation. According to rule 2b, this should still be correct.
        return checkNonRhodf(t, table, store);
    }
    else {
        return true;
    }
}

bool SubpropertyRule::checkReflexive(Triple t, RdfsTable table, RdfsBackend* store)
{
    // If subject in rhodf, return true (6c)
    QList<RhoDF> rhodf = QList<RhoDF>() << RDFS_SP << RDFS_SC << RDFS_TYPE << RDFS_DOM << RDFS_RANGE;
    Q_FOREACH(RhoDF i, rhodf) {
        // Note: map(i) returns empty QString(), which will never be equal to subject
        // which is assumed non-empty; otherwise (_ sp _) is assumed to hold anyway
        if (t.subject == map(i)) {
            return true;
        }
    }

    // Gather triples to check
    QSet<Triple> ts;
    {
        // (X,A,Y)
        Triple tt;
        tt.subject = QString();
        tt.predicate = t.subject;
        tt.object = QString();
        ts << tt;
    }

    // check for (A sp X) or (X sp A)
    Triple t1;
    t1.subject = t.subject;
    t1.predicate = t.predicate;
    t1.object = QString();

    Triple t2;
    t2.subject = QString();
    t2.predicate = t.predicate;
    t2.object = t.subject;
    ts << t1 << t2;

    // And add in dom/range-like terms
    QStringList predicateList;
    if (containsMap(RDFS_DOM)) {
        predicateList << map(RDFS_DOM);
    }
    if (containsMap(RDFS_RANGE)) {
        predicateList << map(RDFS_RANGE);
    }

    Q_FOREACH(QString term, predicateList) {
        Triple tt;
        tt.subject = t.subject;
        tt.predicate = term;
        tt.object = QString();

        ts << tt;
    }

    // Check triples in triple set
    Q_FOREACH(Triple tt, ts) {
        if (store->checkTriple(tt, table) > 0) {
            return true;
        }
    }

    // No match found; return false
    return false;
}

bool SubpropertyRule::checkNonRhodf(Triple t, RdfsTable table, RdfsBackend* store)
{
    // First try to find the triple in G
    if (store->hasTriple(t, table)) {
        return true;
    }

    const QString sp = map(RDFS_SP); // We'll use this several times, so cache it

    const unsigned short maxDistance = (((store->capabilities() & RDFS_TRANSITIVE) > 0) && store->transitivePredicates().contains(t.predicate))
                                     ? store->transitiveDistance()
                                     : 0;

    QQueue<Triple> queue;
    QSet<QString> processed;
    {
        Triple tt;
        tt.subject = QString();
        tt.predicate = sp;
        tt.object = t.predicate;
        queue.enqueue(tt);
        processed << t.predicate;
    }

    Triple it;
    it.subject = t.subject;
    it.predicate = QString(); // Will be filled in later
    it.object = t.object;

    while (!queue.isEmpty()) {
        Triple current = queue.dequeue();

        // Find matches for current triple; note that current.object is the
        // property to look for. incoming.subjects will contain interesting
        // properties to check

        // Note: Munoz et al algorithm first finds triples (A, X, B), then tries
        // to find a path from X to P in G(sp). We go back through G(sp) from P
        // instead, and conclude the triple is in the entailment if we find an
        // (A, X, B) at any time (we then already know (X, sp, P) )

        QList<Triple> incoming = store->findTriples(current, table);
        Q_FOREACH(Triple tt, incoming) {
            // Check that we didn't already process the new property, tt.subject
            if (processed.contains(tt.subject)) {
                continue;
            }

            // Check if we can find (A, X, B) in G
            it.predicate = tt.subject; // A, B already filled in
            if (store->hasTriple(it, table)) {
                return true;
            }

            // Enqueue new subproperty (depending on the distance)
            if (tt.distance >= maxDistance) {
                tt.object = tt.subject;
                tt.subject = QString();
                tt.predicate = sp;
                queue.enqueue(tt);
            }
            processed << tt.object;
        }
    }
    // No match found
    return false;
}


Rdfs::SubpropertyRule::~SubpropertyRule()
{

}
