/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "transitiverule.h"

#include <QStack>
#include <QSet>

#include "../rdfsbackend.h"
#include "../rdfscursor.h"
#include "../rdfslogger.h"

using namespace Rdfs;

Rdfs::TransitiveRule::TransitiveRule()
 : RdfsRule()
{
}

bool TransitiveRule::checkTransitive(Triple t, RdfsTable table, RdfsBackend* store)
{
    // Assume t.predicate is transitive
    QSet<Triple> passed;
    QList<Triple> currentResult;

    const unsigned short maxDistance = (((store->capabilities() & RDFS_TRANSITIVE) > 0) && store->transitivePredicates().contains(t.predicate))
                                     ? store->transitiveDistance()
                                     : 0;

    Triple tf;
    tf.subject = t.subject;
    tf.predicate = t.predicate;
    tf.object = QString();
    passed << tf;

    // Keep at most height of tree 'h' cursors at a time, i.e.
    // construct list with cursor pointers, do DFS through tree
    int i = -1;
    QStack<RdfsCursor *> cursors;
    RdfsCursor *c = store->cursor(tf, table);
    if (c != 0 && c->open()) {
        cursors.push(c);
        i = 0;
    }
    else if (c != 0) {
        delete c;
    }

    // Loop through cursors
    while (i >= 0) {
        tf = c->current();

        // Check if we're a match
        if (tf.object == t.object) {
            // Free open cursors
            qDeleteAll(cursors);
            return true;
        }

        // Only continue search from here if distance >= max transitive distance
        if (tf.distance >= maxDistance) {
            // No match: try opening cursor for child
            tf.subject = tf.object;
            tf.object = QString();
            if (!passed.contains(tf)) {
                passed << tf;

                // Open new cursor (if possible) for subquery
                RdfsCursor *subc = store->cursor(tf, table);
                if (subc != 0 && subc->open()) {
                    c = subc;
                    cursors.push(c);
                    ++i;
                    continue;
                }
                else if (subc != 0) {
                    delete subc;
                }
            }
        }
        // If we got here: item has no subitems, or was already covered earlier
        // I.e.: move to next item, if we reach the end, go back to parent
        c->next();
        while (c->eof()) {
            // No item available anymore - close c and clean it up,
            // go back to parent
            c->close();
            delete c;
            cursors.pop();
            --i;
            if (i >= 0) { // if i < 0 we'll move out of the loop
                c = cursors.top();
                c->next();
            }
            else {
                break;
            }
        }
    }

    // No match found
    return false;
}


Rdfs::TransitiveRule::~TransitiveRule()
{
}
