/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "rdfssql.h"

#include <QObject>
#include <QStringList>
#include <QSqlRecord>
#include <QSqlQuery>
#include <QSqlField>
#include <QVariant>
#include <QDebug>
#include <QSqlError>
#include <QSqlResult>
#include <QCoreApplication>

#include "rdfslogger.h"

using namespace Rdfs;

RdfsSql::RdfsSql(const QString& driver)
    : QSqlDatabase(driver), RdfsBackend()
{
    // Cannot initialize anything
}


RdfsSql::RdfsSql(const QString& driver, const QString& databaseName)
    : QSqlDatabase(QSqlDatabase::addDatabase(driver, "rdfsupdate-" + QString::number(QCoreApplication::applicationPid()))),
      RdfsBackend()
{
    setDatabaseName(databaseName);
    open();

    if (isOpen()) {
        initializeTables();
    }
    else {
        RdfsLogger::instance()->error(QObject::tr("Open error (%1): %2").arg(QString::number(lastError().number()), lastError().text()));
    }
}

bool RdfsSql::isOpen()
{
    return QSqlDatabase::isOpen();
}

void RdfsSql::close()
{
    QSqlDatabase::close();
}

QStringList RdfsSql::databaseTables()
{
    return tables(QSql::Tables);
}

RdfsSql::RdfsSql(const QString& driver, const QString& databaseName,
                             const QString& hostname, const QString& username,
                             const QString& password)
    : QSqlDatabase(QSqlDatabase::addDatabase(driver, "rdfsupdate-" + QString::number(QCoreApplication::applicationPid())))
{
    setDatabaseName(databaseName);
    setHostName(hostname);
    setUserName(username);
    setPassword(password);
    if (open()) {
        initializeTables();
    }
}

int RdfsSql::capabilities() const
{
    return RDFS_NONE;
}


void RdfsSql::initializeTables() {
    initializeRdfsOriginal();
    initializeRdfsRules();
    initializeRdfsUpdate();
}

void RdfsSql::initializeRdfsOriginal() {
    if (!databaseTables().contains("rdfs_original")) {
        QSqlQuery query(*this);
        query.prepare("CREATE TABLE rdfs_original ("
            "id        INTEGER NOT NULL, "
            "subject   VARCHAR(512) NOT NULL, "
            "predicate VARCHAR(512) NOT NULL, "
            "object    VARCHAR(512) NOT NULL, "
            "PRIMARY KEY (id), "
            "INDEX idx_subject   (subject), "
            "INDEX idx_predicate (predicate), "
            "INDEX idx_object    (object), "
            "UNIQUE idx_unique (subject, predicate, object))"
        );
        executeQuery(&query);
    }
}

void RdfsSql::initializeRdfsRules() {
    if (!databaseTables().contains("rdfs_rule_triple")) {
        QSqlQuery query(*this);
        query.prepare("CREATE TABLE rdfs_rule_triple ("
            "id        INTEGER NOT NULL, "
            "subject   VARCHAR(512) NOT NULL, "
            "predicate VARCHAR(512) NOT NULL, "
            "object    VARCHAR(512) NOT NULL, "
            "PRIMARY KEY (id), "
            "INDEX idx_subject   (subject), "
            "INDEX idx_predicate (predicate), "
            "INDEX idx_object    (object), "
            "UNIQUE idx_unique (subject, predicate, object))"
            );
        executeQuery(&query);
    }

    if (!databaseTables().contains("rdfs_rule")) {
        QSqlQuery query(*this);
        query.prepare("CREATE TABLE rdfs_rule ("
            "source    INTEGER NOT NULL, "
            "target    INTEGER NOT NULL, "
            "PRIMARY KEY (source, target), "
            "INDEX idx_source    (source), "
            "INDEX idx_target    (target))");
        executeQuery(&query);
    }
}

void RdfsSql::initializeRdfsUpdate() {
    if (!databaseTables().contains("rdfs_update")) {
        QSqlQuery query(*this);
        query.prepare("CREATE TABLE rdfs_update ("
            "id        INTEGER NOT NULL, "
            "subject   VARCHAR(512) NOT NULL, "
            "predicate VARCHAR(512) NOT NULL, "
            "object    VARCHAR(512) NOT NULL, "
            "PRIMARY KEY (id), "
            "INDEX idx_subject   (subject), "
            "INDEX idx_predicate (predicate), "
            "INDEX idx_object    (object), "
            "UNIQUE idx_unique (subject, predicate, object))"
        );
        executeQuery(&query);
    }
}

int RdfsSql::tripleId(const Rdfs::Triple& t, RdfsTable table)
{
    QSqlQuery query(*this);
    query.prepare("SELECT id FROM " + tableName(table) + " WHERE subject = ? AND predicate = ? AND object = ?");
    query.addBindValue(t.subject);
    query.addBindValue(t.predicate);
    query.addBindValue(t.object);
    executeQuery(&query);
    if (query.next()) {
        bool ok;
        int result;
        result = query.record().field("id").value().toInt(&ok);
        if (ok) {
            return result;
        }
        else {
            return -1;
        }
    }
    else {
        return -1;
    }
}

void RdfsSql::insertTriple(const Triple &t, RdfsTable table) {
    // Triple has to be valid
    if (t.subject.isEmpty() || t.predicate.isEmpty() || t.object.isEmpty()) {
        return;
    }

    // Variables
    QString tableStr = tableName(table);
    QSqlField field;
    int id = 0;
    bool ok; // Result of converting record value to integer

    // Start transaction
    transaction();
    QSqlQuery query(*this);

    // Get next index
    query.prepare("SELECT MAX(id) FROM " + tableStr);
    executeQuery(&query);
    query.next();
    field = query.record().field(0);
    id = field.isNull() ? 1 : field.value().toInt(&ok) + 1;
    if (!ok) {
        id = 1;
    }

    // Insert record
    query.prepare("INSERT INTO " + tableStr + " (id, subject, predicate, object) "
                  "VALUES (?, ?, ?, ?)");
    query.bindValue(0, id);
    query.bindValue(1, t.subject);
    query.bindValue(2, t.predicate);
    query.bindValue(3, t.object);
    executeQuery(&query);
    commit();
}

int RdfsSql::checkTriple(const Rdfs::Triple& t, RdfsTable table) {
    QString tableStr = tableName(table);

    // Build query
    QString queryString = "SELECT COUNT(*) FROM " + tableStr;
    QStringList conditions;
    QStringList bindValues;
    if (!t.subject.isEmpty()) {
        conditions << "subject = ?";
        bindValues << t.subject;
    }

    if (!t.predicate.isEmpty()) {
        conditions << "predicate = ?";
        bindValues << t.predicate;
    }

    if (!t.object.isEmpty()) {
        conditions << "object = ?";
        bindValues << t.object;
    }

    if (!conditions.isEmpty()) {
        queryString += " WHERE " + conditions.join(" AND ");
    }

    // Prepare query and bind values
    QString value;
    QSqlQuery query(*this);
    query.prepare(queryString);
    Q_FOREACH(value, bindValues) {
        query.addBindValue(value);
    }

    // Execute and get value
    executeQuery(&query);
    query.next();
    QSqlField field = query.record().field(0);
    bool ok;
    int count = field.isNull() ? 0 : field.value().toInt(&ok);
    if (!ok) {
        count = 0;
    }
    return count;
}

void RdfsSql::removeTriple(const Triple &t, RdfsTable table) {
    // Triple has to be valid
    if (t.subject.isEmpty() || t.predicate.isEmpty() || t.object.isEmpty()) {
        return;
    }

    // Variables
    QString tableStr = tableName(table);
    QSqlField field;

    QSqlQuery query(*this);

    // Insert record
    query.prepare("DELETE FROM " + tableStr + " WHERE "
                  "subject = ? AND "
                  "predicate  = ? AND "
                  "object = ?");
    query.bindValue(0, t.subject);
    query.bindValue(1, t.predicate);
    query.bindValue(2, t.object);
    executeQuery(&query);
}

void RdfsSql::truncate (RdfsTable table)
{
    QString tableStr = tableName(table);

    QSqlQuery query(*this);
    query.prepare("TRUNCATE TABLE " + tableStr);
    executeQuery(&query);
}

QList<Triple> RdfsSql::findTriples(const Rdfs::Triple& t, RdfsTable table) {
    QString tableStr = tableName(table);

    // Build query
    QString queryString = "SELECT id, subject, predicate, object FROM " + tableStr;
    QStringList conditions;
    QStringList bindValues;
    if (!t.subject.isEmpty()) {
        conditions << "subject = ?";
        bindValues << t.subject;
    }

    if (!t.predicate.isEmpty()) {
        conditions << "predicate = ?";
        bindValues << t.predicate;
    }

    if (!t.object.isEmpty()) {
        conditions << "object = ?";
        bindValues << t.object;
    }

    if (!conditions.isEmpty()) {
        queryString += " WHERE " + conditions.join(" AND ");
    }

    // Prepare query and bind values
    QString value;
    QSqlQuery query(*this);
    query.prepare(queryString);
    Q_FOREACH(value, bindValues) {
        query.addBindValue(value);
    }

    // Execute built query
    executeQuery(&query);

    // Get triples
    QList<Triple> result;
    #ifndef NDEBUG
    int i = 0;
    #endif
    while (query.next()) {
        Triple t;
        t.subject = query.record().field("subject").value().toString();
        t.predicate = query.record().field("predicate").value().toString();
        t.object = query.record().field("object").value().toString();
        result.append(t);

        #ifndef NDEBUG
        if (i && (i % 1000) == 0) {
            RdfsLogger::instance()->log(QObject::tr("Read %1 rows").arg(QString::number(i)));
        }
        ++i;
        #endif
    }
    return result;
}

void RdfsSql::insertRule(Rule r) {
    // Insert triples in rule db, if not already done
    const RdfsTable table = RDFS_RULE;
    if (checkTriple(r.first, table) == 0) {
        insertTriple(r.first, table);
    }
    if (checkTriple(r.second, table) == 0) {
        insertTriple(r.second, table);
    }

    // Get ID numbers of triples
    int t1id, t2id;
    t1id = tripleId(r.first, table);
    t2id = tripleId(r.second, table);

    if (t1id == -1 || t2id == -1) {
        return; // Invalid triples
    }

    // Insert rule
    QSqlQuery insertQuery(*this);
    insertQuery.prepare("INSERT INTO rdfs_rule (source, target) VALUES (?, ?)");
    insertQuery.addBindValue(t1id);
    insertQuery.addBindValue(t2id);
    executeQuery(&insertQuery);
}

QList<Rule> RdfsSql::rules() {
    QList<Rule> result;
    const QString table = tableName(RDFS_RULE);
    QSqlQuery query(*this);
    query.prepare("SELECT s.subject ssubject, s.predicate spredicate, s.object sobject, "
                  "t.subject tsubject, t.predicate tpredicate, t.object tobject "
                  "FROM rdfs_rule r LEFT JOIN " + table + " s ON r.source = s.id LEFT JOIN "
                  + table + " t ON r.target = t.id");
    executeQuery(&query);
    while (query.next()) {
        Triple t1, t2;
        t1.subject = query.record().field("ssubject").value().toString();
        t1.predicate = query.record().field("spredicate").value().toString();
        t1.object = query.record().field("sobject").value().toString();

        t2.subject = query.record().field("tsubject").value().toString();
        t2.predicate = query.record().field("tpredicate").value().toString();
        t2.object = query.record().field("tobject").value().toString();

        result << Rule(t1, t2);
    }
    return result;
}

void RdfsSql::removeRule(Rule r) {
    const RdfsTable table = RDFS_RULE;
    // Retrieve the triple ids
    QMap<int, Triple> map;
    int tid1, tid2, rid;

    transaction();

    tid1 = tripleId(r.first, table);
    tid2 = tripleId(r.second, table);

    if (tid1 == -1 || tid2 == -1) {
        return; // Invalid triples -> invalid rule
    }

    // Delete the rule
    QSqlQuery query(*this);
    query.prepare("DELETE FROM rdfs_rule WHERE source = ? AND target = ?");
    query.addBindValue(tid1);
    query.addBindValue(tid2);
    executeQuery(&query);

    // Clean up the two triples if they're no longer used
    QSqlQuery cquery1(*this);
    cquery1.prepare("SELECT COUNT(*) FROM rdfs_rule WHERE source = ? OR target = ?");
    // Add twice, for both bindings
    cquery1.addBindValue(tid1);
    cquery1.addBindValue(tid1);
    executeQuery(&cquery1);
    cquery1.next();
    RdfsLogger::instance()->debug(cquery1.record().field(0).value().toString());
    if (cquery1.record().field(0).value().toInt() == 0) {
        removeTriple(r.first, table);
    }

    QSqlQuery cquery2(*this);
    cquery2.prepare("SELECT COUNT(*) FROM rdfs_rule WHERE source = ? OR target = ?");
    // Add twice, for both bindings
    cquery2.addBindValue(tid2);
    cquery2.addBindValue(tid2);
    executeQuery(&cquery2);
    cquery2.next();
    RdfsLogger::instance()->debug(cquery2.record().field(0).value().toString());
    if (cquery2.record().field(0).value().toInt() == 0) {
        removeTriple(r.second, table);
    }
    commit();
}

void RdfsSql::executeQuery(QSqlQuery *query) {
    query->exec();
    #ifndef NDEBUG
    RdfsLogger::instance()->debug(QObject::tr("Executed: %1").arg(query->executedQuery()));
    // Build string list of bound values
    QStringList params;
    for (int i = 0; i < query->boundValues().size(); ++i) {
        params << query->boundValue(i).toString();
    }
    if (params.count() > 0) {
        RdfsLogger::instance()->debug(QObject::tr("Parameters: %1").arg(params.join(", ")));
    }
    #endif

    QSqlError error = query->lastError();
    switch(error.type()) {
        case QSqlError::NoError:
            break;
        case QSqlError::ConnectionError:
            RdfsLogger::instance()->error(QObject::tr("Connection error (%1): %2").arg(QString::number(error.number()), error.text()));
            break;
        case QSqlError::StatementError:
            RdfsLogger::instance()->error(QObject::tr("Statement error (%1): %2").arg(QString::number(error.number()), error.text()));
            break;
        case QSqlError::TransactionError:
            RdfsLogger::instance()->error(QObject::tr("Transaction error (%1): %2").arg(QString::number(error.number()), error.text()));
            break;
        case QSqlError::UnknownError:
        default:
            RdfsLogger::instance()->error(QObject::tr("Unknown error (%1): %2").arg(QString::number(error.number()), error.text()));
            break;
    };
}

Rdfs::RdfsSql::~RdfsSql() {
    this->close();
}
