/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef RDFSSQL_H
#define RDFSSQL_H

#include <QtSql/QSqlDatabase>
#include <QtSql/QSqlQuery>

#include "rdfsbackend.h"

namespace Rdfs {

/**
 * This class interfaces with a database managing the RDFS triples in a database.
 *
 * A database consists of three tables:
 * - "rdfs_original": Original RDFS Graph
 * - "rdfs_rules": RDFS Rules
 * - "rdfs_update": Update graphs
 *
 * The default implementation runs basic queries which should work for most DBMSes.
 * If a query does not work, create a subclass and override the necessary queries.
 *
 * Use executeQuery() to execute a specific query on the database. This will call
 * the query's exec() function, and log errors et cetera.
 *
 * To execute multiple queries as part of a transaction, call transaction() and
 * commit() first.
 */
class RdfsSql : public QSqlDatabase, public RdfsBackend {
    protected:
        /**
         * Execute a single query and (if debugging is enabled) log any output
         * to the debug output.
         * @param query Pointer to the query to execute
         */
        void executeQuery(QSqlQuery *query);

        /**
         * Check if a database connection is established
         */
        void initializeTables();

        /**
         * Constructor. Initializes the database, but does not connect anything.
         * Use this constructor when deriving this class and you overriding any
         * of the initialize* functions; the virtuals tables
         * won't be available yet, so the initialization will not call your
         * functions, but the base implementations.
         *
         * Make sure your constructor(s) opens a database connection, and if
         * successful calls initializeTables().
         *
         * @param driver The driver to use
         */
        explicit RdfsSql(const QString &driver);

        /**
         * Create the rdfs_original table if it doesn't exist yet
         */
        virtual void initializeRdfsOriginal();
        /**
         * Create the rdfs_rules table if it doesn't exist yet
         */
        virtual void initializeRdfsRules();
        /**
         * Create the rdfs_udpate table if it doesn't exist yet.
         */
        virtual void initializeRdfsUpdate();

        /**
         * Get a list of tables in the current database. The default implementation
         * simply calls tables(QSql::Tables) and returns this. For drivers which
         * support accessing multiple databases with the same connection, this
         * may need to be overridden. Otherwise, this may return table names of
         * tables in all databases the user can access.
         */
        virtual QStringList databaseTables();

        /**
         * Get the triple's unqiue ID in the given table.
         * @param t The triple to find
         * @param table The table to look in
         * @return The numerical ID of the triple, or -1 if it was not found
         */
        virtual int tripleId(const Triple &t, RdfsTable table = RDFS_ORIGINAL);

    public:
        /**
         * Constructor. Connect to a database and create the necessary tables.
         * If the connection fails, the database is (naturally) not initialized.
         *
         * @param driver The driver to use
         * @param databasename The database to use
         */
        RdfsSql(const QString &driver, const QString &databaseName);

        /**
         * Constructor. Connect to a database and create the necessary tables.
         * If the connection fails, the database is (naturally) not initialized.
         *
         * @param driver The driver to use
         * @param databasename The database to use
         * @param hostname The hostname to connect to (if applicable)
         * @param username The username to use to connect to the database (if applicable)
         * @param password The password to use to connect to the database.
         */
        RdfsSql(const QString &driver, const QString &databaseName,
                const QString &hostname, const QString &username,
                const QString &password);

        /**
         * Destructor
         */
        virtual ~RdfsSql();

        virtual bool isOpen();

        virtual void close();

        /**
         * Get the capabilites of this backend.
         */
        virtual int capabilities() const;

        /**
         * Insert a new triple in the database
         * @param t The triple to insert. Must be a complete triple (no blanks).
         * @param table The table to check in (Original or Update)
         */
        virtual void insertTriple(const Triple &t, RdfsTable table = RDFS_ORIGINAL);

        /**
         * Check how many rows match the given triple. If a triple field (subject,
         * predicate or object) is empty, it is not checked on. An empty triple
         * (all empty strings) gives the amount of total rows in the given table.
         * @param t The triple to check (may contain blanks).
         * @param table The table to check in (Original or Update)
         */
        virtual int checkTriple(const Triple &t, RdfsTable table = RDFS_ORIGINAL);

        /**
         * Find the triples which match the given triple @p t (with optionally blank
         * subject, predicate and/or object, which will then be ignored in the query).
         * The triples are returned in a map, with their id (unique within the table)
         * as key.
         * @param t The triple to check (may contain blanks).
         * @param table The table to check in (Original or Update)
         */
        virtual QList<Triple> findTriples(const Rdfs::Triple& t, RdfsTable table);

        /**
         * Remove a triple from the database
         * @param t The triple to remove. Must be a complete triple (no blanks).
         * @param table The table to check in (Original or Update)
         */
        virtual void removeTriple(const Triple &t, RdfsTable table = RDFS_ORIGINAL);

        /**
         * Insert a new rule. The two triples must be identified by Id, as given by
         * the findTriplesMap() function map (in the key), on the RDFS_RULE table.
         * @param r The rule to insert
         */
        virtual void insertRule(Rule r);
        using RdfsBackend::insertRule;

        /**
         * Retrieve all deterministic tuple-generating dependencies in a list.
         */
        virtual QList<Rule> rules();

        /**
         * Remove a rule by its triples. If the given triples do not form a defined
         * rule, nothing happens.
         * @param r The rule (triple pair) to remove.
         */
        virtual void removeRule(Rule r);
        using RdfsBackend::removeRule;

        /**
         * Truncate a table. This removes all triples in the given table.
         * @param table The table to clear (Original or Update).
         */
        virtual void truncate(RdfsTable table = RDFS_ORIGINAL);
};
};
#endif // RDFSSQL_H
