/*
    RdfsUpdate: A research tool for RDFS data
    Copyright (C) 2009-2011 Eindhoven University of Technology - www.tue.nl
    Copyright (C) 2009-2011 Wouter Haffmans <wouter@simply-life.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef RDFS_RDFSTRIPLETBACKEND_H
#define RDFS_RDFSTRIPLETBACKEND_H

#include "rdfsbackend.h"

#include <QStringList>

#include "rdfscursor.h"

namespace Rdfs {

typedef QPair<QString, unsigned char> TertiaryAtom;

/**
 * A TripleT payload.
 *
 * This contains three buckets: subject, predicate and object, which define
 * the role of the atom of the key. Each bucket is a key/value pair, designating
 * the secondary item with a list of tertiary items.
 */
struct TripleTPayload {
    // TODO QMap (sorted by key) or QHash (hash-table lookup)? Perhaps compare?
    // Probably insignificant difference - we're interested in I/O :)

    /**
     * Does the payload contain data, or is the data located in a separate B-tree?
     *
     * For normal payloads this will be true, with the data set in the
     * TripleTPayload::data member. For popular atoms, this will be false.
     * To retrieve/set triples for that
     */
    bool hasData;

    /**
     * Map from secondary -> tertiary index. If the key's role is the subject,
     * this is Object -> Predicates. If the key's role is predicate, this is
     * Subject -> Objects. For the object role, this is Subject -> Predicate.
     *
     * Note that the integer depicts the distance of the triples, in case of
     * transitive predicates
     */
    QMap<QString, QList< TertiaryAtom > > data;
};

/**
 * A base class for Triple-T indexed back-ends. This base takes care of the
 * serializing/deserializing of the payload to a byte array, which can be
 * stored in the actual backend.
 *
 * A subclass needs to provide methods to load and put the contents
 * (TripleT payload) of an atom. Additionally, it needs to provide a list
 * of all atoms available in the index, for completely open queries (no atom
 * given).
 *
 * Other methods that still need to be implemented are opening/closing the
 * database and isOpen(). Also, the truncate() method needs to be implemented.
 * If the backend provides statistics, the stats() method also needs to be
 * implemented.
 *
 */
class RdfsTripleTBackend : public Rdfs::RdfsBackend
{

private:
    friend class RdfsTripleTCursor;

    bool m_inBulkLoad;
    quint64 m_bulkLoadCount;
    static const quint64 m_maxBulkLoadSize;
    static const uint m_maxPrimaryPayloadPages;
    static const QStringList m_transitivePredicates;
    int m_maxPrimaryPayloadSize;
    unsigned char m_distance;
    QString m_settingsPath;
    QHash<QString, TripleTPayload> m_bulkPayloadsOriginal;
    QHash<QString, TripleTPayload> m_bulkPayloadsUpdate;

    /**
     * Guess payload data size based on a payload object
     */
    static quint32 payloadSize(TripleTPayload payload);

    void insertSorted(TertiaryAtom item, QList< TertiaryAtom >* list);

    /**
     * Check the payload size and store it if it's suitable for the primary index. If it is
     * too large, clear it and store all contained triples in the secondary BTrees
     */
    void storePayload(const QString &primary, const RdfsRole &role, const TripleTPayload &triples, const RdfsTable &table);

    /**
     * Use binary search to find the position of a tertiary item, in a list. This returns
     * an iterator at the position of the item, or list->end() if the item is not found.
     */
    static QList<TertiaryAtom>::iterator binaryFindTertiary(QList<TertiaryAtom> *list, const QString &tertiary);

    /**
     * Directly insert a triple into the data store (non-bulk mode). This will also insert
     * necessary inferred triples, depending on the distance.
     * @param t The triple to insert
     * @param table The table to insert the triple in
     */
    void insertNonbulkTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table = RDFS_ORIGINAL);

    /**
     * Insert a single triple directly into the data store, without checking for inferred
     * triples etc.
     * @param t The triple to insert
     * @param table The table to insert the triple in
     */
    void insertSingleTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table = RDFS_ORIGINAL);

    /**
     * Insert a triple in bulk mode. Include inferred triples which need to be stored as well,
     * depending on the distance setting.
     * @param t The triple to insert
     * @param table The table to insert the triple in
     */
    virtual void insertBulkTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table = RDFS_ORIGINAL);

    /**
     * Insert a single triple in bulk mode, without inferred triples.
     * @param t The triple to insert
     * @param table The table to insert the triple in
     */
    virtual void insertSingleBulkTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table = RDFS_ORIGINAL);

    /**
     * Find inferred triples which need to be materialized after adding a triples with a
     * transitive predicate. This triple doesn't need to be materialized yet. This finds
     * all triples, i.e. those on the left of the given triple, on the right and those where
     * the given triple is in the middle.
     *
     * @param t The triple to find the inferred triples for (a, p, b)
     * @param table The table to search in.
     * @param additionalLeft Additional triples on the left side of the @p t (?, p, b)
     * @param additionalRight Additional triples on the right side of the @p t (a, p, ?)
     */
    QList<Triple> findInferredTriples(const Rdfs::Triple& t, Rdfs::RdfsTable table = RDFS_ORIGINAL,
                                      const QList<Triple> &additionalLeft = QList<Triple>(),
                                      const QList<Triple> &additionalRight = QList<Triple>());

    /**
     * Find inferred triples which need to be materialized after adding a triples with a
     * transitive predicate. This triple doesn't need to be materialized yet. This finds
     * the triples either on the "left" side or the "right" side of the given triple, depending
     * on the search. Use findTriples() to get both including those inferred triples,
     * where the original triple is in the middle of the chain.
     *
     * Note that in the @p search triple, either the subject or object must be empty (not both). The empty
     * one is filled in in the result, which will be based on the @p original triple. The distance
     * of the original triple and found triples will never exceed the m_distance setting.
     *
     * @param search The query triple to find inferred triples.
     * @param original The original triple which is being inserted
     * @param table The table in which the triple is inserted, and where triples are looked up in.
     */
    QList<Triple> findSubInferredTriples(const Rdfs::Triple& search, const Rdfs::Triple& original, Rdfs::RdfsTable table = RDFS_ORIGINAL);

    static TripleTPayload deserializePayload(const QByteArray &payload);
    static QByteArray serializePayload(TripleTPayload payload);
    TripleTPayload mergePayloads(const Rdfs::TripleTPayload& a, const Rdfs::TripleTPayload& b) const;
    QList< TertiaryAtom > mergeLists(const QList< Rdfs::TertiaryAtom >& a, const QList< Rdfs::TertiaryAtom >& b) const;

    /**
     * Check if a predicate is transitive or not
     */
    bool isTransitive(const QString &predicate) const;

    /**
     * Remove a triple from the store, with inferrence. This will recursively remove
     * triples such that the entire chain (with @p t as starting point) is removed.
     * @param t The triple to remove. This must have the distance set properly!
     * @param table The table to remove the triples from.
     */
    void removeInferredTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table = RDFS_ORIGINAL);

    /**
     * Remove a complete triple (one without blanks).
     *
     * The normal removeTriple() method first finds all triples which match using the index,
     * then removes them appropriately using this method.
     *
     * Otherwise, if one of the atoms is blank, one entire sub-index would have to be traversed
     * in order to find all matching triples and to remove them.
     */
    void removeCompleteTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table = RDFS_ORIGINAL);

#ifndef NDEBUG
    bool m_savingStatistics, m_loadedStatistics;
    ulong m_tripleCount, m_dataSize;
    void loadStatistics();
    void saveStatistics();
    void recomputeStatistics(const QString &atom, const RdfsRole &role, const TripleTPayload &payload);
    uint payloadTripleCount(const Rdfs::TripleTPayload& payload);
#endif

    // TODO Add options:
    // - QMap/QHash lookups (see TripleTPayload struct)
    // - Max count of items in bulk "cache" (m_maxBulkLoadSize)
    // - Spread data (payload) over max. x (uncompressed) bytes per item
    //   - i.e. split up data of heavily-used atoms, such as "type"

protected:
    /**
     * Load the (serialized) payload for a specific atom and the role it's in.
     * If the atom does not exist yet for the given role, the payload should be
     * an empty QByteArray.
     */
    virtual QByteArray payload(const QString &atom, const RdfsRole &role, Rdfs::RdfsTable table = RDFS_ORIGINAL) = 0;
    /**
     * Store the (serialized) payload for a specific atom in a given role.
     *
     * The given payload may be empty (QByteArray::isEmpty()). In that case it is safe to remove the
     * payload from the database entirely.
     */
    virtual void putPayload(const QString &atom, const RdfsRole &role, QByteArray data, Rdfs::RdfsTable table = RDFS_ORIGINAL) = 0;
    /**
     * Get a list of all atoms in the graph.
     */
    virtual QStringList atoms(Rdfs::RdfsTable table = RDFS_ORIGINAL) = 0;

    /**
     * Store a triple a in a separate B-tree. The B-tree to use is determined by the @p atom, @p role and @p table. The
     * @p secondary and @p tertiary strings define the other two atoms in the triple, in the order in which
     * they should be indexed.
     *
     * Note that the implementation does not necessarily have to implement a B-tree, although that is most likely
     * the more efficient method to retrieve the data.
     *
     * Also note that both during search and saving, the roles of the secondary and tertiary items are undefined.
     */
    virtual void insertBtreeTriple(const QString &atom, const RdfsRole &role, const QString &secondary, const QString &tertiary, unsigned char distance, Rdfs::RdfsTable table) = 0;

    /**
     * Remove a triple from an additional B-tree database. The B-tree to use is determined by the @p atom and @p role. The
     * @p secondary and @p tertiary strings define the other two atoms in the triple, in the order in which
     * they should be indexed. Note that at any time, all three atoms will be given.
     */
    virtual void removeBtreeTriple(const QString &atom, const RdfsRole &role, const QString &secondary, const QString &tertiary, Rdfs::RdfsTable table) = 0;

    /**
     * Get a cursor which selects items in a specific B-tree.
     *
     * The B-tree to use is determined by the @p atom, @p role and @p table.
     *
     * Note that secondary and/or tertiary may be empty, in which case multiple triples need to be returned (with
     * the empty items matching anything).
     *
     * The result is a list of triples. The exact role of the secondary and tertiary items is generally unknown,
     * so the results should be as follows:
     * - Subject: the given @p atom
     * - Predicate: atoms found in the @p secondary role
     * - Object: atoms found in the @p tertiary role
     * The caller will have to manipulate these triples as required.
     */
    virtual RdfsCursor *btreeCursor(const QString &atom, const RdfsRole &role, const QString &secondary, const QString &tertiary, Rdfs::RdfsTable table) = 0;

    /**
     * Set the page size of the storage back-end. This will determine how large payloads can be before
     * data is stored in a secondary B-tree index.
     *
     * By default the page size is 4096 bytes (4 Kb). The B-tree index will be used as soon as
     * a payload exceeds 4 pages.
     *
     * @param size The page size. This must be at least 4.
     */
    void setPageSize(uint size);

public:
    RdfsTripleTBackend(QString settingsPath);
    virtual ~RdfsTripleTBackend();

    virtual QStringList options() const { return QStringList() << "distance"; };
    virtual bool setOption(const QString& key, const QString& value);

    virtual void startBulkLoad();
    virtual void endBulkLoad();
    virtual Rdfs::RdfsCursor* cursor(Rdfs::Triple q, Rdfs::RdfsTable table = RDFS_ORIGINAL);
    virtual bool hasTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table = RDFS_ORIGINAL);
    virtual void removeTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table = RDFS_ORIGINAL);
    virtual QList< Rdfs::Triple > findTriples(const Rdfs::Triple& t, Rdfs::RdfsTable table);
    virtual int checkTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table = RDFS_ORIGINAL);
    virtual void insertTriple(const Rdfs::Triple& t, Rdfs::RdfsTable table = RDFS_ORIGINAL);

    virtual void insertRule(Rule r);
    virtual void removeRule(Rule r);
    virtual QList< Rule > rules();

    virtual int capabilities() const;
    virtual unsigned char transitiveDistance() const;
    virtual QStringList transitivePredicates() const;



};

}

#endif // RDFS_RDFSTRIPLETBACKEND_H
